/** JOYCE PCW emulator for DOS/DJGPP *************************/
/**                                                         **/
/** JOYCE.C: contains the main PCW specific code            **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1999                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/


/*
 * Note: JOYCE may well contain assumptions that an int is 32 bits or more.
 * I wouldn't recommend porting it to a 16-bit system - the ability
 * to do a malloc(2megs) is quite important :-)
 *
 */

#define MAIN
#include "Joyce.h"
#include "Joycefid.h"
#include "Joyceptr.h"
#include "Joyceset.h"
#include "Joycedos.h"
#include "Joycemou.h"
#include "Joycelgo.h"
#include "Joycedsk.h"
#include "ay8910.h"
#include "inline.h"

void StartCPM(void);
void TrashCPM(void);
void LoadBoot(void);
void UnitJoyce(void);

/* Argument processing */

static int autoboot = -1;

void checkarg(char *s)
{
	if (!strcasecmp(s,"-DEBUG")) DebugMode = 1;
	if ((s[1] == 'B' || s[1] == 'b') && s[2] == '=')
	{
		drive_filename[1] = s + 3;
		printf("Set drive B: to %s\n", drive_filename[1]);
	}
	/* v1.22 Alternative bootfile (undocumented) */ 
	if ((s[1] == 'E' || s[1] == 'e') && s[2] == '=')
	{
		strcpy(bootfile, s + 3);
		printf("Set boot ROM to %s\n", bootfile);
	}
	/* v1.33 Automatic boot (still undocumented) */
	if ((s[1] == 'A' || s[1] == 'a') && s[2] == '=')
	{
		autoboot = atoi(s + 3);
		printf("Automatically booting from boot disc %d\n", autoboot); 
	}
}


/* This function is called once every cycle to check for a reboot 
  (forcing reload of the boot image and hardware reset) and for the
  termination of an FDC operation (jstate.fdcint == 1).

*/

void JoyceZ80(Z80 *R)
{
	if (jstate.reset_flag)
	{
		/* Zero the memory mapped keyboard */
		
		vga_setmode(0);
		memset(&PCWRAM[0xFFF0],0,16);

		/* set Program Counter to 0000H */
		jstate.reset_flag = 1;
		ResetZ80(R);

		/* Page in the bottom 64k of memory */

		mem_page(0xF0,0x80);
		mem_page(0xF1,0x81);
		mem_page(0xF2,0x82);
		mem_page(0xF3,0x83);
		mem_lock(0xF1);
		LoadBoot();	/* Reload the boot ROM */

		jstate.timeron = 1; 
		jstate.timerint = 1;
		jstate.reset_flag = 0;		
		jstate.mouse_patched = 0; /* (v1.27) reset mouse variables */
		jstate.md3_mx = 0;
		jstate.md3_my = 0;
		jstate.md3_mf1 = 0;
		jstate.md3_mf2 = 0;
		jstate.sp_mx = 0;
		jstate.sp_my = 0;	
		mouse_reset();
	}

	/* Check for end of FDC transfer */
	if (jstate.fdcint == 1)
	{
		jstate.inf8 |= 0x20;	/* FDC is interrupting */
		jstate.fdcint = 0;
		switch (jstate.fdc_int_mode)
		{
			case 2: IntZ80(R, 0x66);
				break;

			case 3: IntZ80(R, 0x38);
				break;

			default: jstate.fdcint = 0;
				 break;
		}
	}
	else if (jstate.fdcint > 1) --jstate.fdcint;
}


/* Reload the boot ROM */

void LoadBoot(void)
{	/* Open the boot ROM */

	FILE *fp;
	word address = 0;
	int c;

	if ((fp = fopen(bootfile,"rb")) == NULL) {
		UnitScr();
		fprintf(stderr,"Open of %s failed!  \r\n", bootfile);
		diewith("Failed to open boot file\r\n",98);
	}
	/* Load the boot ROM into Z80 memory */
	while ((c = getc(fp)) != EOF) PCWRAM[address++] = c;

	fclose(fp);
}



/******************* main() *******************/

int init_joyce(int argc, char **argv)
{
	int na;

	joycename=argv[0]+strlen(argv[0]);
	joycedir=".";

	while (--joycename != argv[0])
	{
		if (*joycename=='\\' || *joycename=='/' || *joycename == ':')
		{
			*joycename=0;
			joycename++;
			joycedir=argv[0];
			break;
		}
	}

        /* load default boot image (v1.22: moved to before checkarg() call) */
        strcpy(bootfile, find_boot_file(BOOT_SYS));

        for (na = 1; na < argc; na++)
        {
                if (argv[na][0] == '-') checkarg(argv[na]);
        }

        fplog = fopen(find_joyce_file(LOGFILE),"w");

        if (!read_settings())
        {
                if(jset.verbose) printf("Using default settings\n");
                default_settings();
        }
	return 0;
}





#if defined(__MSDOS__) || defined(USE_SDL)
	/* WX JOYCE uses the wxApp class as its main() */


void main(int argc,char *argv[])
{
        printf("JOYCE PCW Emulator v%x.%x\n", BCDVERS >> 8, BCDVERS & 0xFF);

	init_joyce(argc, argv);

	if (jset.verbose) printf("Joyce directory is %s\n\r",joycedir);

	/* trap signals so the terminal doesn't go to hell
	 * when we exit
	 */
	signal(SIGINT, goodbye);

	atexit(UnitJoyce);

	com_init();

	/* kick off CP/M */
	StartCPM();

	/* Return memory */
	TrashCPM();

	/* we should exit through the goodbye() function
	 * to restore terminal modes and close disk drives
	 */
	goodbye(99);

	/* The logic will never get here */
}


#endif /* def __MSDOS__ || def USE_SDL */

/* Initialise the emulator */

void StartCPM(void)
{
	word A;
	int *T;
	int memsiz;
	int c;

	(void)c;

	/*** STARTUP CODE starts here: ***/

	T=(int *)"\01\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
#ifdef LSB_FIRST
	if (*T!=1) {
		fprintf(stderr, "********** This machine is high-endian. **********\r\n");
		fprintf(stderr, "Take #define LSB_FIRST out and compile JOYCE again.\r\n");
		fflush(stderr);
		goodbye(99);
	}
#else
	if (*T==1) {
		fprintf(stderr, "********* This machine is low-endian. **********\r\n");
		fprintf(stderr, "Insert #define LSB_FIRST and compile JOYCE again.\r\n");
		goodbye(99);
	}
#endif

/* Set initial values */
	jstate.inf8 = 0x30;
	jstate.fdcint = 0;
	jstate.timerint = 1;
	jstate.timeron = 1;	//0;
	jstate.reset_flag = 0;

#ifdef __MSDOS__
/* Allocate 800x600 memory for a screen buffer. */

	Framebuf=malloc(800*600);
	if (!Framebuf) memsiz = 0; 
	else memsiz = mem_alloc();
#else	/* __MSDOS__ */
	memsiz = mem_alloc();
#endif
	if (!memsiz)
	{
#ifdef __MSDOS__
		if(Framebuf) free (Framebuf);
#endif
		fprintf(stderr,"Sorry... Not enough memory to run JOYCE.\r\n");
		fflush(stderr);
		goodbye(99);
	}
	else if (jset.verbose)	/* Say how much memory we got */
		printf("This is a PCW 8%d\r\n",memsiz);

#ifdef __MSDOS__	/* Don't bother under UNIX */
	cputs("Press SPACE to continue...\r\n");
	c=getch();
#endif

	InitPtr();	  /* Initialise the printer */
	InitFDC(0);	  /* Initialise the emulated FDC */
	InitAdlib();
	if (!InitScr(0))  /* Initialise the emulated screen */
	{
		TrashAdlib();
		fprintf(stderr,"Sorry, could not select 800x600 graphics mode.\n");
		fprintf(stderr,"%s needs at least 800x600 8-bit graphics.\n",joycename);
		exit(2);
	}
	InitMouse();	  /* Initialise the mouse */

	/* All "outside functions" are regulated in JOYCE by a
	  simulated 900Hz clock. For a 4MHz CPU, this is every
	  4e6/9e2 = 4444 cycles, or about 500 instructions. 
          On a PCW with a Sprinter, double this; but if the interrupts get 
          too infrequent, they will never occur when EI is on. I'm 
          using 4441 rather than 4444 because it's prime, and so we're 
          less likely to get problems caused by interrupts synchronising 
          exactly with a tight loop. */

	jstate.Rg.IPeriod=4441;  /* Every 4441 cycles (~900Hz) check if
			  something needs doing */

	/* Initialise the emulated keyboard */
	if (!InitKbd()) diewith("DPMI error\r\n",98);

	LoadBoot();

	extern_stop = 0;	/* v1.31 */

	/* set Program Counter to 0000H */
	ResetZ80(&jstate.Rg);
	jstate.Rg.Trace = 0;
	jstate.Rg.Trap = 0xFFFF;

	/* Page in the bottom 64k of memory */

	mem_page(0xF0,0x80);
	mem_page(0xF1,0x81);
	mem_page(0xF2,0x82);
	mem_page(0xF3,0x83);

	if (autoboot > 0 && fdc_boot(autoboot))
	{
		jstate.Rg.PC.W = 0xF010;
		jstate.Rg.SP.W = 0xF000;
		jstate.Rg.IFF = 0;
	}


	/* Finally... set the Z80 going... */

	A=RunZ80(&jstate.Rg);

	/*if(jset.verbose)*/ cprintf("EXITED at PC = %Xh.\r\n",A);

}



/****************************************************************/
/*** Extended JOYCE instructions. This is called on the ED FD ***/
/*** instruction to bypass anything in the BIOS that needs    ***/
/*** bypassing. There are no register parameters; the 	      ***/
/*** function is calculated from the address of the call.     ***/
/****************************************************************/

void BIOSZ80(Z80 *R)
{
	/* (v1.26) Support for printer diversion */

	if (R->PC.W < 0xFC12 && R->PC.W >= 0xFC0F) /* LIST */
	{
		list(R);
	}
	if (R->PC.W < 0xFC30 && R->PC.W >= 0xFC2D) /* LISTST */
	{
		listst(R);
	}
	/* (v1.26) ends */

	if (R->PC.W < 0xFC51 && R->PC.W >= 0xFC4E) /* TIME */
	{
		word temp = R->DE.W;

		R->DE.W = 0xFBF4;
		jdos_time(R, 1);

		R->DE.W = temp;
	}
}




/****************************************************************/
/*** Extended JOYCE instructions. This is called on the ED FE ***/
/*** instruction to emulate anything that distinguishes JOYCE ***/
/*** from a real PCW					      ***/
/****************************************************************/

void PatchZ80(Z80 *R)
{
	switch(R->AF.B.h)	/* Register A */
	{
		case 0x00:	/* Get JOYCE version */
			R->AF.B.h = 0xFF;	/* Utility compatibility number. 
						 * 0  for a real Z80
						 * -1 for Joyce 1.00+ */
			R->HL.W = BCDVERS;	/* JOYCE version */
			R->DE.W = 'J';		/* Emulator ID */
			break;

		case 0x01:
/*
       A=1:  Boot. If B=0, read sector 1 from floppy drive & enter it;
		   If B>0, read sector 1 from image .\BOOT\BOOTn.DSK and
			   enter it.
*/
			if (fdc_boot(R->BC.B.h))
			{
				R->PC.W = 0xF010;
				R->SP.W = 0xF000;
				R->IFF = 0;
			}
			break;

		case 0x02: 
/*
	A=2: Set colour. BCD = 'white' RGB
			 EHL = 'black' RGB
*/
			screen_colour( R->BC.B.h, R->BC.B.l, R->DE.B.h,
				       R->DE.B.l, R->HL.B.h, R->HL.B.l);
			break;

/* 
	A=3 - A=7: Service a call from a disc driver FID

*/
		case 0x03:
			fid_d_login(R);
			break;
		case 0x04:
			fid_d_read(R);
			break;
		case 0x05:
			fid_d_write(R);
			break;
		case 0x06:
			fid_d_flush(R);
			break;
		case 0x07:
			fid_ems(R);
			break;
		case 0x08:		/* A=8: Save settings */
			write_settings(); 			
			break;
		case 0x09:		/* A=9: Switch which LPT we're using */
			R->HL.W = jset.cen_port;
			if (R->DE.W != 0xFFFF) jset.cen_port = R->DE.W;
			break;
		case 0x0A:		/* A=10: Access DOS filesystem */
			jdos(R);
			break;
		case 0x0B:
			fid_char(R);	/* A=11: Screen control FID */
			break;
		case 0x0C:
			KeyboardED(R);	/* A=12: low-level keyboard mapping */
			break;
		case 0x0D:
			fid_com(R);	/* (v1.21) A=13: COMPORT.FID */
			break;
		case 0x0E:
			fid_d_mess(R);	/* (v1.22) FID disc messages */
			break;
		case 0x0F:		/* (v1.22) LOGO */
			dr_logo(R);
			break;
		case 0x10:		/* (v1.22) parse BOOT/BOOT.CFG */
			boot_id(R);
			break;
		case 0xFC:		/* (v1.30) CPMREDIR interface */
			cpmredir_intercept(R);
			break;
		case 0xFD:		/* (v1.27) MD3 mouse patch */
			mouse_patch(R);
			break;
		case 0xFE:		/* Set JOYCE timing parameters */
			R->HL.W   = R->IPeriod;
			R->AF.B.h = jset.screen_refresh;
			if (R->BC.B.l) jset.screen_refresh = R->BC.B.l;
			if (R->DE.W) R->IPeriod = R->DE.W;
			break;
		case 0xFF:		/* Halt & catch fire */
			goodbye(R->HL.W);
			break;

		default:		/* Invalid ED xx */
			joyce_dprintf("Unknown operation: ED FE with A=%02x at PC=%04x",
				R->AF.B.h, R->PC.W);
			break;
	}
	return;
}


/****************************************************************/
/*** Write byte to given IO port.			      ***/
/****************************************************************/
void OutZ80(register word Port,register byte Value)
{
	switch(Port & 0xFF)
	{
		case 0: /* Floppy disc controller */
		case 1: FDCport(Port,Value); break;

		case 0x84: /* CEN: */
		case 0x85:
		case 0x86:
		case 0x87: cen_out(Port, Value); 

		case 0xA0: /* AY-8910 */
		case 0xA1:
		case 0xA2:
		case 0xA3: AYPort(Port, Value);

		case 0xE1: /* (v1.24) These are serial port accesses,   */
		case 0xE2: /*        but I'm using them as a convenient */ 
		case 0xE3: /*        way of starting the timer. Sadly   */
		case 0xE4: /*        this doesn't work for Loco 1       */
		case 0xE5: /*        because it doesn't use the SIO     */
		case 0xE7: /*  */
		jstate.timeron= 1;
		break;

		case 0xF0: /* Memory paging */
		case 0xF1:
		case 0xF2:
		case 0xF3: mem_page(Port,Value); break;

		case 0xF4: mem_lock(Value); break; 

		case 0xF5: /* Video controller */
		case 0xF6:
		case 0xF7: VidPort(Port,Value); break;

		case 0xF8: /* Odds and ends */
			   switch(Value)
		{
			case 1: /* Reboot! */
			jstate.reset_flag = 1;
			break;

			/* Zero the memory mapped keyboard */

			memset(&PCWRAM[0xFFF0],0,16);

			/* set Program Counter to 0000H */
			jstate.reset_flag = 1;
			jstate.Rg.PC.W = 0x0000;

			/* Page in the bottom 64k of memory */

			mem_page(0xF0,0x80);
			mem_page(0xF1,0x81);
			mem_page(0xF2,0x82);
			mem_page(0xF3,0x83);
			mem_lock(0xF1);
			LoadBoot();	/* Reload the boot ROM */

			break;

			case 2: /* Change the way FDC interrupts */
			case 3:
			case 4:
			jstate.fdc_int_mode = Value;
			break;

			case 5:	/* Set the FDC terminal count */
			case 6:	/* Setting it aborts data transfer */
			jstate.fdc_terminal_count = Value;
			fdc_set_tcount(Value);
			break;

			case 7:
			screen_on();
			break;

			case 8:
			screen_off();
			break;

			case 9: /* FDC motors on */
			fdc_mstart();
			break;

			case 10: /* FDC motors off */
			fdc_mstop();
			break;

			case 11: /* Beeper on */
			beep_on();
			break;

			case 12: /* Beeper off */
			beep_off();
			break;

			default:
			break;
		}
		break;

		case 0xFC:	/* Printer control, ignored */
		case 0xFD:	/* (v1.20) */
		break;

		default:
		joyce_dprintf("OUT 0x%02x, 0x%02x\n", Port, Value);
		break;
	}
}

/****************************************************************/
/*** Read byte from given IO port.			      ***/
/****************************************************************/
byte InZ80(register word Port)
{
	byte v;

	switch (Port & 0xFF)
	{
		case 0: /* Floppy disc controller */
		case 1:
		return(FDCin(Port));
		break;

		case 0x84: /* CEN: */
		case 0x85:
		case 0x86:
		case 0x87: return cen_in(Port & 0xFF); 

		case 0xF4:
		v = jstate.inf8;
		jstate.inf8 &= 0xF0;	/* Reset 300Hz counter */
		return v;

		case 0xF8:	/* Odds & ends */
		return(jstate.inf8);
		break;

		case 0xFC:	/* (v1.20) Printer */
		return 0xF8;	/* No printer */
		break;

		case 0xFD:	/* (v1.20) Printer */
		if (jstate.model_type == 1) return 0xFD; /* 9512 */
		return 0xC8;	/* No printer */
		break;

		case 0xA0:	/* AMX mouse */
		case 0xA1:
		case 0xA2:
		case 0xD0:	/* Kempston mouse */
		case 0xD1:
		case 0xD2:
		case 0xD3:
		case 0xD4:
		return MouseIn(Port & 0xFF);
		break;
	}
	return(-1);
}


/****************************************************************/
/*** This function is called at 900Hz. It has to implement    ***/
/*** all PCW features which rely on time.		      ***/
/****************************************************************/
word LoopZ80(Z80 *R)
{
	static int Tcount=0, Scount=0, Rcount=0;

	if (extern_stop) return INT_QUIT;

	{
		vga_timer();
		Tcount++;		/* Tickers used to drop the 900Hz */
		Scount++;		/* down to 50Hz & 300Hz */
		Rcount++;

		if (Scount >= jset.screen_refresh) /* Draw the screen at 50Hz */
		{
			DrawScr();	   /* Redraw the screen */
			Scount=0;
		}
		if (Rcount == 18) Rcount = 0;
		if (Rcount < 3) 	       /* In the vertical retrace */
			jstate.inf8 |= 0x40;
		else	jstate.inf8 &= 0xBF;

		PollKbd(0, R);	/* Poll the keyboard */

		if (Tcount >= 3) /* 300Hz signal */
		{
			Tcount = 0;
			if ((jstate.inf8 & 0x0F) < 0x0F) jstate.inf8++;	/* Increase the 300Hz counter */
			if (jstate.timeron || jstate.timerint)
			{
				jstate.timerint = 0;
				if (R->IFF & 0x01)     /* EI */ return 0x0038;
				R->IRequest = 0x0038;  /* Pending INT */
				return 0x0038;
			}
		}
	}
	return 0xFFFF;	/* No interrupt */
}

/* Free memory */

void TrashCPM(void)
{
	/* de allocate memory for Z80 CPU */
	if (PCWRAM) free(PCWRAM);
	PCWRAM=NULL;
#ifdef __MSDOS__
	if (Framebuf) free(Framebuf);
	Framebuf=NULL;
#endif
}


/* signal trap - reset tty modes and leave program
 * 99 is used if the program wants to exit gracefully,
 */
void goodbye(int sig)
{
	if (fplog)
	{
		if (sig != 99) fprintf(fplog, "Stopped by signal: %d\n", sig);
		fclose(fplog);
	}
	if(PCWRAM) free(PCWRAM);
#ifdef __MSDOS__
	if(Framebuf) free(Framebuf);
#endif
	fid_cleanup();
	UnitKbd();
	UnitPtr();	/* (v1.11) */
	UnitFDC();
	UnitMouse();
        UnitScr();
	TrashAdlib();
	exit(sig);
}


/* As goodbye(), but prints a farewell message */

void diewith(char *s, int sig)
{
	if (fplog)
	{
		fprintf(fplog, "Fatal error %d: %s", sig, s);
		fclose(fplog);
	}
	if(PCWRAM) free(PCWRAM);
#ifdef __MSDOS__
	if(Framebuf) free(Framebuf);
#endif
	fid_cleanup();
	UnitPtr();	/* (v1.11) */
	UnitKbd();
	UnitFDC();
	UnitScr();
	UnitMouse();
	TrashAdlib();
	cputs(s);
	cputs("Press SPACE to leave...\n\r");
	getch();
	exit(sig);
}



void UnitJoyce(void)	/* v1.21 The very last things done */
{
	fid_fd_unit();	/* v1.22 */
	com_deinit();
}
