/** JOYCE PCW emulator for Linux/SDL *************************/
/**                                                         **/
/** JoyceSdlKbd.c: contains the SDL interface               **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1999                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/


#include "Joyce.h"

volatile char keyboard_map[17];
volatile char keyboard_flag;

/* NOTE: This won't compile out of the box on SDL 0.8.10. You need to add
  the keys SDLK_HASH, SDLK_LSUPER and SDLK_RSUPER 

  Changes in v1.35 for SDL 0.10.0
*/

word sdl_keyboard_xlt[] = { 
	SDLK_BACKSPACE, 9, 128, 9, 128,
	SDLK_TAB,       8, 16,  8, 16,
//	SDLK_CLEAR,
	SDLK_RETURN,    2, 4,   2, 4, 
	SDLK_ESCAPE,    1, 1,   1, 1,
	SDLK_SPACE,     5, 128, 5, 128,
	SDLK_QUOTE,     3, 16,  3, 16,
	SDLK_COMMA,     4, 128, 4, 128,
	SDLK_MINUS,     3, 2,   3, 2,
	SDLK_PERIOD,    3, 128, 3, 128,
	SDLK_SLASH,     3, 64,  3, 64,
	SDLK_0,         4, 1,   4, 1,
	SDLK_1,         8, 1,   8, 1,
	SDLK_2,         8, 2,   8, 2,
 	SDLK_3,         7, 2,   7, 2,
	SDLK_4,         7, 1,   7, 1,
	SDLK_5,         6, 2,   6, 2,
	SDLK_6,         6, 1,   6, 1,
	SDLK_7,         5, 2,   5, 2,
	SDLK_8,         5, 1,   5, 1,
	SDLK_9,         4, 2,   4, 2,
	SDLK_BACKSLASH, 2, 64,  2, 64,
	SDLK_SEMICOLON, 3, 32,  3, 32,
	SDLK_EQUALS,    3, 1,   3, 1,
	SDLK_KP0,       0, 2,   0, 2,
	SDLK_KP1,       2, 16,  1, 128,
	SDLK_KP2,       10,64,  0, 128,
	SDLK_KP3,       0, 16,  0, 64,
	SDLK_KP4,       1,128,  1, 32,
	SDLK_KP5,       0,128,  1, 64,
	SDLK_KP6,       0, 64,  0, 32,
	SDLK_KP7,       1, 32,  2, 16,
	SDLK_KP8,       1, 64,  1, 16,
	SDLK_KP9,       0, 32,  0, 16,
	SDLK_KP_PERIOD, 10, 4, 10, 64,
	SDLK_KP_DIVIDE,  3,64,  3, 64,
	SDLK_KP_MULTIPLY,1, 2,  1,  2,
	SDLK_KP_MINUS,   1,128, 1, 32,
	SDLK_KP_PLUS,    2,128, 2,128,
	SDLK_KP_ENTER,   10,32,10, 32,
//	SDLK_KP_EQUALS = 86,
	SDLK_LEFTBRACKET,3,  4, 3,  4,
	SDLK_HASH,        2, 8, 2,  8,
	SDLK_RIGHTBRACKET,2, 2, 2,  2,
	SDLK_BACKQUOTE,   8, 4, 8,  4,
	SDLK_a,           8,32, 8, 32,
	SDLK_b,           6,64, 6, 64,
	SDLK_c,          7, 64, 7, 64,
	SDLK_d,          7, 32, 7, 32,
	SDLK_e,          7,  4, 7,  4,
	SDLK_f,          6, 32, 6, 32,
	SDLK_g,          6, 16, 6, 16,
	SDLK_h,          5, 16, 5, 16,
	SDLK_i,          4,  8, 4,  8,
	SDLK_j,          5, 32, 5, 32,
	SDLK_k,          4, 32, 4, 32,
	SDLK_l,          4, 16, 4, 16,
	SDLK_m,          4, 64, 4, 64,
	SDLK_n,          5, 64, 5, 64,
	SDLK_o,          4,  4, 4,  4,
	SDLK_p,          3,  8, 3,  8,
	SDLK_q,          8,  8, 8,  8,
	SDLK_r,          6,  4, 6,  4,
	SDLK_s,          7, 16, 7, 16,
	SDLK_t,          6,  8, 6,  8,
	SDLK_u,          5,  4, 5,  4,
	SDLK_v,          6,128, 6,128,
	SDLK_w,          7,  8, 7,  8,
	SDLK_x,          7,128, 7,128,
	SDLK_y,          5,  8, 5,  8,
	SDLK_z,          8,128, 8,128,
	SDLK_DELETE,     2,  1, 2,  1,
	SDLK_F1,         0,  4, 0,  4,
	SDLK_F2,         0,  4, 0,  4,
	SDLK_F3,         0,  1, 0,  1,
	SDLK_F4,         0,  1, 0,  1,
	SDLK_F5,        10,  1,10,  1,
	SDLK_F6,        10,  1,10,  1,
	SDLK_F7,        10, 16,10, 16,
	SDLK_F8,        10, 16,10, 16,
	SDLK_F9,        16,  1,16,  1,	// Joyce actions
	SDLK_F10,       16,  2,16,  2,	// Joyce exit
	SDLK_F11,        2,128, 2,128,
	SDLK_F12,       10,  8,10,  8,
	SDLK_F13,       10,  2,10,  2,
	SDLK_F14,       10,  2,10,  2,
	SDLK_F15,       16,  1,16,  1,
	SDLK_PAUSE,     16,  1,16,  1,
	SDLK_NUMLOCK,    0,  2, 0,  2,
	SDLK_UP,         1, 64, 1, 64,
	SDLK_DOWN,      10, 64,10, 64,
	SDLK_RIGHT,      0, 64, 0, 64,
	SDLK_LEFT,       1,128, 1,128,
	SDLK_INSERT,     1,  4, 1,  4,
	SDLK_HOME,       1,  8, 1,  8,
	SDLK_END,        9,  8, 9,  8,
	SDLK_PAGEUP,     0,  8, 0,  8,
	SDLK_PAGEDOWN,   1, 16, 1, 16,
	SDLK_CAPSLOCK,   10,32,10, 32,
	SDLK_SCROLLOCK,  8, 64, 8, 64,
	SDLK_RSHIFT,     2, 32, 2, 32,
	SDLK_LSHIFT,     2, 32, 2, 32,
	SDLK_RCTRL,      10, 2, 10, 2,
	SDLK_LCTRL,      10, 2, 10, 2,
	SDLK_RALT,       10,128,10,128,
	SDLK_LALT,       10,128,10,128,
	SDLK_RMETA,      10,128,10,128,
	SDLK_LMETA,      10,128,10,128,
//	SDLK_HELP = 164,
	SDLK_PRINT,      1, 2, 1, 2,  
	SDLK_SYSREQ,     1, 2, 1, 2,
	SDLK_MENU,       16,1,16, 1,
	SDLK_BREAK,      16,2,16, 2,
//	SDLK_EURO,     
	SDLK_POWER,      16,2,16,2, 
	SDLK_LSUPER,     10,2,10,2,
	SDLK_RSUPER,     10,2,10,2,
	0
};

static word keyb_map[4 * SDLK_LAST];

SDL_Event ev;

// Event loop...
void eventloop(void)
{
	static int aushift = 0, trushift = 0;
	int evid = SDL_PollEvent(&ev);
	word *km;

	while (evid)
	{
		if (ev.type == SDL_KEYDOWN)
		{
			if (ev.key.keysym.sym == SDLK_F2 ||
                            ev.key.keysym.sym == SDLK_F4 ||
                            ev.key.keysym.sym == SDLK_F6 ||
                            ev.key.keysym.sym == SDLK_F8)  aushift = 1; 

			if (ev.key.keysym.sym == SDLK_LSHIFT ||
                            ev.key.keysym.sym == SDLK_RSHIFT) trushift = 1;
		}

		if (ev.type == SDL_KEYUP)
		{
                        if (ev.key.keysym.sym == SDLK_F2 ||
                            ev.key.keysym.sym == SDLK_F4 ||
                            ev.key.keysym.sym == SDLK_F6 ||
                            ev.key.keysym.sym == SDLK_F8)  aushift = 0;

			if (ev.key.keysym.sym == SDLK_LSHIFT ||
                            ev.key.keysym.sym == SDLK_RSHIFT) trushift = 0;
		}

		switch(ev.type)
		{

			case SDL_QUIT: 
				keyboard_map[0x10] |= 2; 
				break; /* Quit key */
			case SDL_KEYDOWN:
				km = keyb_map + 4 * ev.key.keysym.sym;
				keyboard_map[km[0]] |= km[1];
				keyboard_flag = 1;
				break;
			case SDL_KEYUP:
				km = keyb_map + 4 * ev.key.keysym.sym;
				keyboard_map[km[0]] &= ~km[1];
				keyboard_flag = 1;
				break;
			case SDL_MOUSEMOTION:
		                if (tracking) track_mouse(ev.motion.x, ev.motion.y);
				break;
		}
	switch(ev.type)
	{
		case SDL_KEYUP:
		case SDL_KEYDOWN:
		if (trushift || aushift) keyboard_map[2] |= 32;
		else		 	 keyboard_map[2] &= ~32;
		break;
	}
	evid = SDL_PollEvent(&ev);
	}	
}

static void rebuild_map(void)
{
	int n;
	word *b;
	
	memset(keyb_map, 0, sizeof(keyb_map));
	
	for (b = sdl_keyboard_xlt; *b; b += 5)
	{
		n = b[0];
		keyb_map[n*4]   = b[1];		
		keyb_map[n*4+1] = b[2];		
		keyb_map[n*4+2] = b[3];		
		keyb_map[n*4+3] = b[4];
	}		
}

int InitKbd(void) 
{
	rebuild_map();
	KeySet();
	return 1;
}              

void UnitKbd() { }


void keyboard_chain(int toggle) 
{
//        SDL_EventState(SDL_KEYDOWN, SDL_IGNORE);
//        SDL_EventState(SDL_KEYUP, SDL_IGNORE);
}


void AssignKey(word keynum, byte l, byte h)
{
	int n = 4 * keynum;
	keyb_map[n]   = l;
	keyb_map[n+1] = h;
	keyb_map[n+2] = l;
	keyb_map[n+3] = h;
}

void KeyboardED(Z80 *R)
{
	int n, l, h, k;

	switch(R->BC.B.l)
	{
		case 0:
		R->HL.B.l = keyb_map[R->DE.W * 4];
		R->HL.B.h = keyb_map[R->DE.W * 4 + 1];
		break;

		case 1:
		AssignKey(R->DE.W, R->HL.B.l, R->HL.B.h);
		break;
		
		case 2:
		for (n = 0; n < R->DE.W; n++)
		{
			k =       RdZ80(R->IX.W + 4*n)
			  | 256 * RdZ80(R->IX.W + 4*n + 1);
			l =       RdZ80(R->IX.W + 4*n + 2);
			h =       RdZ80(R->IX.W + 4*n + 3);
			
			AssignKey(k, l, h);	
		} 
		break;
	}
}


void PollKbd(char mutex, Z80 *R)
{
        unsigned char keyno;

	eventloop();

        if (!keyboard_flag) return;
        keyboard_flag=0;

        for (keyno=0;keyno<16;keyno++)
        {
                PCWRAM[0xFFF0+keyno]=keyboard_map[keyno];
                PCWRAM[0xFFFD] |= 0x80;
        }

        if ((!mutex) && (keyboard_map[0x10]&1))
        {
                JoyceMenu(R);
                for (keyno=0;keyno<16;keyno++)
                {
                        PCWRAM[0xFFF0+keyno]=0;
                }
                return;
        }
        if (keyboard_map[0x10]&2)
        {
                ExitConf();
                for (keyno=0;keyno<16;keyno++)
                {
                        PCWRAM[0xFFF0+keyno]=0;
                }
                return;
        }
}


static int ALTS[] = { SDLK_LALT,  SDLK_RALT, SDLK_LMETA, SDLK_RMETA, -1 };
static int CTLS[] = { SDLK_LCTRL, SDLK_RCTRL, -1 };
static int BKSP[] = { SDLK_BACKSPACE, -1 };
static int DELS[] = { SDLK_DELETE,    -1 };

static void setRange(int *R, byte a, byte b, byte c, byte d)
{
	int n;

        for (; R[0] >= 0; ++R)
        {
                n = *R; 
                keyb_map[n*4]   = a;
                keyb_map[n*4+1] = b;
                keyb_map[n*4+2] = c;
                keyb_map[n*4+3] = d;
        }

}

void KeySet(void)
{
	if (jset.swap_alt)
	{
		setRange(ALTS, 10,   2, 10,   2);
		setRange(CTLS, 10, 128, 10, 128);
	}
	else
	{
                setRange(ALTS, 10, 128, 10, 128);
                setRange(CTLS, 10,   2, 10,   2);
	}
	if (jset.swap_delete)
	{
		setRange(BKSP, 2,   1, 2,   1);
		setRange(DELS, 9, 128, 9, 128);
	}
	else
	{
		setRange(BKSP, 9, 128, 9, 128);
		setRange(DELS, 2,   1, 2,   1);
	}
}


unsigned char get_keytoken(void)
{
        int evid = SDL_WaitEvent(&ev);

        while (1)
        {
		evid = SDL_WaitEvent(&ev);
                if (ev.type == SDL_KEYDOWN) return ev.key.keysym.sym;
		if (ev.type == SDL_QUIT)    return SDLK_y;
	}
}

