/** JOYCE PCW emulator for SDL *******************************/
/**                                                         **/
/** (v1.22) Screen access for Digital Research Logo         **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1999                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "Joycelgo.h"
#include "Joycevga.h"
#include "inline.h"

#define P1_WORD RdZ80(R->SP.W + 2) + 256*RdZ80(R->SP.W + 3)
#define P2_WORD RdZ80(R->SP.W + 4) + 256*RdZ80(R->SP.W + 5)
#define P3_WORD RdZ80(R->SP.W + 6) + 256*RdZ80(R->SP.W + 7)
#define P4_WORD RdZ80(R->SP.W + 8) + 256*RdZ80(R->SP.W + 9)

#define P1_BYTE RdZ80(R->SP.W + 2)
#define P3_BYTE RdZ80(R->SP.W + 6)	/* After 2 words */
#define P4_BYTE RdZ80(R->SP.W + 7)
#define P5_BYTE RdZ80(R->SP.W + 10)	/* After 4 words */
#define P6_BYTE RdZ80(R->SP.W + 11)

#define ESC "\0x1B"

static void vgastr(char *s)
{
	while (*s)
	{
		vga_putchar(*s);
		++s;
	}
}


/* LIOS interface:
 
  E=function number
 SP->stack frame:
	DW	return		or	DW	return
	DW	param1			DB	param1 
	DW	param2			DB	param2
	...				...

functions are:

E=	Parameters 	Description
--------------------------------------------------------------------
0:	none		Initialise. By this time, the info block has
			already been used to set up the screen size;
			so the Amstrad LIOS (and the JOYCE LIOS) 
			change it when the RSX is awakened (function 
			0FFh below) and then restart the COM file 
			with a JP 100h. This routine, which will be
			called the second time around, is therefore
			a no-op.
1:	word 0		Terminate
2:	none		Turn on text mode, no graphics
3:	byte line	Set split line (no. of text lines visible)
4:	none		CONIN - return in HL
5:	none		CONST - return in HL
6:	byte ch		CONOUT - print char if text screen available
7:	word T,		Clear text window from T to B
	word B
8:	word X,		Move text cursor to X,Y
	word Y
9:	word T,		Scroll text window from T to B
	word B
10:	word X,		Plot a point. pen is 2 for PX else PD
	word Y,		
	byte pen,
	byte colour
11:	word X1,	Draw a line from X1,Y1 to X2,Y2
	word Y1,
	word X2,
	word Y2,
	byte pen,
	byte colour
12:	none?		no-op?
13:	word colour,	Clear screen - height=pixels
	word pixels
14:	none		Beep
15:	none		Get top of memory
16:	word X,		Read pixel colour into L
	word Y
17:	none		LISTST
18:	byte ch		LIST
19-23:	none		no-ops? Not called by DR Logo
24:	word BUF,	Generate record <RECNO> of a picture savefile.
	word RECNO	Return 1 in HL if more records follow
25:	word BUF,	Load record <RECNO> of a picture savefile.
	word RECNO

26:	none?		These are not hooked in the Amstrad LIOS or JOYCE, but
27:	none?		are used for 'pal' and 'setpal' respectively.

0FFh: 	word PARAMS	Initialise. Enter with HL->message buffer, write 
			message here if there is a failure. Return HL=0
			if failure, else 1

PARAMS:	DW	0	;+0	Text screen height, chars
	DW	0	;+2	Text screen width, chars
	DW	0	;+4	Text screen height, chars
	DW	0	;+6	Text screen width, chars
	DW	0	;+8	Standard split line, chars
	DW	0	;+A	Default foreground colour
	DW	0	;+C	Default background colour
	DW	0	;+E
	DW	0	;+10	Number of foreground colours - 1
	DW	0	;+12
	DW	0	;+14
	DW	0	;+16
	DW	0	;+18	Graphics screen width, pixels
	DW	0	;+1A	Graphics screen height, pixels
	DW	0	;+1C
	DW	0	;+1E	Maximum scrunch ratio
	DW	0	;+20	1 if split screen disabled?
	DW	0	;+22	Number of background colours - 1
	DW	0	;+24	Text screen height, chars
	DW	0	;+26	Scrunch ratio, numerator
	DW	0	;+28	Scrunch ratio, denominator
	DW	0	;+2A	->Version-specific title string, ASCIIZ
	DW	0	;+2C	->"Edit", ASCIIZ
	DW	0	;+2E	->"_",    ASCIIZ

*/

static byte oldtype;
static int splitsz, gfx_y;


static void dosplit(void)
{
	if (!splitsz) gfx_y = 600;
	else gfx_y = 16 * splitsz; 
	GrSetClipBox(0, 0, 800, 600);
}

void __inline__ WrByte(word a, byte v)
{
	WrZ80(a,     v);
	WrZ80(a + 1, 0);
}




int makerop(byte pen, byte *colour)
{
	switch(pen)
	{
		default: return 0;		/* pd */
		case 2:  *colour = 0; return 2;	/* pu */
		case 3:	 return 3;		/* px */
		case 4:  *colour = 0; return 0;	/* pe */
	}
}

void lios0(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios4(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios5(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios12(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios17(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios18(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios19(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios20(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios21(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios22(void) {}	/* Possible no-ops, functions implemented in Z80 */
void lios23(void) {}	/* Possible no-ops, functions implemented in Z80 */

void zstrcpy(word dest, char *src)
{
	do
	{
		WrZ80(dest++, *src);
	} while (*src++);
}

word lios_init(word liospb, word msgbuf)
{
	char title[90];

	sprintf(title,"Amstrad LOGO v2.0 under JOYCE %x.%02x", BCDVERS >> 8, BCDVERS & 0xFF);
	oldtype = ScrType;
	vgastr(" ");			/* Force it into 800x600 VGA */
	vga_cursor(0);
	vga_bg = colour_find(0,0,0);		/* Black */
	vga_fg = colour_find(255,255,255);	/* White */
	GrClearScreen(vga_bg);
	vga_curvis = 0;
	WrByte(liospb,      36);
	WrByte(liospb + 2,  99);
	WrByte(liospb + 4,  36);
	WrByte(liospb + 6,  99);
	WrByte(liospb + 8,  27);
	WrWord(liospb +10, (word)vga_fg);
	WrWord(liospb +12, (word)vga_bg);
	WrWord(liospb +16, 255);		/* Allow 256 colours */
	WrWord(liospb +24, 800);
	WrWord(liospb +26, 600);
	WrWord(liospb +34, 255);		/* 256 background colours */
	WrByte(liospb +36,  36);
	WrWord(liospb +38,  32);		/* 1:1 scrunch ratio */
	WrWord(liospb +42,  msgbuf);
	zstrcpy(msgbuf,title); 
	splitsz = 27;
	dosplit();
        SDL_UpdateRect(screen, 0, 0, 0, 0);
	return 1;
}

void lios1(void)
{
	vga_fg = dwhi;
	vga_bg = dblk;
	vgastr(ESC "1" ESC "w");
	vga_cursor(1);
	ScrType = oldtype;
}

void lios2(void)
{
	splitsz = -1;
	vga_curon = 1;
}

void lios3(byte split)
{
	splitsz = split;
	if (!splitsz)	/* All graphics no text */
		vga_curon = 0;
	else	vga_curon = 1;
	dosplit();
}

void __inline__ lios6(byte ch)
{
	if (splitsz) vga_putchar(ch);	/* Don't do text in fullscreen */
}

void lios7(word T, word B)	/* Clear text lines T to B */
{
	long top    = (T-1) * 16;
	long bottom = (B-1) * 16 + 15;

	if (bottom > 599) bottom = 599;
	if (top > 599)    top    = 599;

	if (vga_curvis) 
	{
		vga_drawcur(); 
		vga_curvis = 0;
	}
	GrFilledBox(0, top, 799, bottom, colour_find(0,0,0));
        SDL_UpdateRect(screen, 0, top, 800, bottom-top);

}

void lios8(word X, word Y)  /* Move text cursor to X,Y */
{
	vga_putchar(0x1B);
	vga_putchar('Y');
	vga_putchar(Y+31);
	vga_putchar(X+31);
}


void lios9(word T, word B) /* Scroll text window */
{
	long top    = (T-1) * 16;
	long bottom = (B-1) * 16 + 15;

	if (bottom > 599) bottom = 599;
	if (top > 599)    top    = 599;

	if (vga_curvis) 
	{
		vga_drawcur(); 
		vga_curvis = 0;
	}
	SdlScrBlit(0, top, 0, top + 16, 799, bottom);
	GrFilledBox(0, bottom - 16, 799, bottom, colour_find(0,0,0));
        SDL_UpdateRect(screen, 0, top, 800, bottom-top);
}

word lios10(word X, word Y, byte colour, byte pen)
{
	int rop = makerop((pen == 2) ? 3 : 1, &colour);	

	GrPlotRop(X, Y, colour, rop);
        SDL_UpdateRect(screen, X, Y, 1, 1);
	return 1;
}


word lios11(word X1, word Y1, word X2, word Y2, byte colour, byte pen)
{
	int rop = makerop((pen == 2) ? 3 : 1, &colour);	

        GrLineRop(X1, Y1, X2, Y2, colour, rop);

	if (X1 > X2) { word t = X2; X2 = X1; X1 = t; }
	if (Y1 > Y2) { word t = Y2; Y2 = Y1; Y1 = t; }

        SDL_UpdateRect(screen, X1, Y1, X2-X1+1, Y2-Y1+1);
	return 1;
}

word __inline__ lios13(word colour, word pixels)
{
	if (vga_curvis) 
	{
		vga_drawcur(); 
		vga_curvis = 0;
	}
	GrFilledBox(0, 0, 799, pixels, colour);
        SDL_UpdateRect(screen, 0, 0, 0, 0);
	return 1;
}

void __inline__ lios14(void)
{
	vga_putchar(7);
}


word lios15(void)
{
	return RdZ80(6)+256*RdZ80(7);
}

word __inline__ lios16(word X, word Y)
{
	return (word)(GrPixel(X, Y));
}


word lios24(word buf, word recno)
{
	return get_bmp_record(buf, recno);
}

word lios25(word buf, word recno)
{
	return put_bmp_record(buf, recno);
}

void dr_logo(Z80 *R)
{
/*	joyce_dprintf("LIOS call %d ", R->DE.B.l);
	switch(R->DE.B.l)
	{
		case 3: joyce_dprintf("(%d)", P1_BYTE); break;
		case 6: joyce_dprintf("(%c)", P1_BYTE); break;
		case 7:
		case 8:
		case 9:
		case 13:
		case 16:joyce_dprintf("(%d,%d)", P1_WORD, P2_WORD); break;
		case 10:joyce_dprintf("(%d,%d,%d,%d)", P1_WORD, P2_WORD, P3_BYTE, P4_BYTE); break;
		case 11:joyce_dprintf("(%d,%d,%d,%d,%d,%d)", P1_WORD, P2_WORD, P3_WORD, P4_WORD, P5_BYTE, P6_BYTE); break;
	}
	joyce_dprintf(" SP=%x\n", R->SP.W);
*/	switch(R->DE.B.l)
	{
		case 0:	lios0(); break;
		case 1: lios1(); break;
		case 2: lios2(); break;
		case 3:	lios3(P1_BYTE); break;
		case 4: lios4(); break; /* done in Z80 code */
		case 5: lios5(); break;	/* done in Z80 code */
		case 6: lios6(P1_BYTE); break;
		case 7: lios7(P1_WORD, P2_WORD); break;
		case 8: lios8(P1_WORD, P2_WORD); break;
		case 9: lios9(P1_WORD, P2_WORD); break;
		case 10:R->HL.W = lios10(P1_WORD, P2_WORD, P3_BYTE, P4_BYTE); break;
		case 11:R->HL.W = lios11(P1_WORD, P2_WORD, P3_WORD, P4_WORD,
			       P5_BYTE, P6_BYTE); break;
		case 12:lios12(); break;
		case 13:R->HL.W = lios13(P1_WORD, P2_WORD); break;
		case 14:lios14(); break;
		case 15:R->HL.W = lios15(); break;
		case 16:R->HL.W = lios16(P1_WORD, P2_WORD); break;
		case 17:lios17(); break; /* done in Z80 code */
		case 18:lios18(); break; /* done in Z80 code */
		case 19:lios19(); break;
		case 20:lios20(); break;
		case 21:lios21(); break;
		case 22:lios22(); break;
		case 23:lios23(); break;
		case 24:R->HL.W = lios24(P1_WORD, P2_WORD); break;
		case 25:R->HL.W = lios25(P1_WORD, P2_WORD); break;

		case 0xFF: R->HL.W = lios_init(P1_WORD, R->HL.W); break;

		default: joyce_dprintf("Unrecognised LIOS call %d\n", R->DE.B.l);
	}
}

/* BMP headers. HDR1 used for I/O; stdhdr holds an 800x600x256 bmp header */
	
byte hdr1[54];
byte stdhdr[54] = {'B', 'M', 	  /* (2) Magic number */
		   54, 87, 7, 0,  /* (4) file size */
		   0,   0, 0, 0,  /* (4) zero */
		   54,  4, 0, 0,  /* (4) offset to image data */
		   40,  0, 0, 0,  /* (4) length of info area*/
		   32, 3, 0, 0,	  /* (4) width  = 800 */
		   88, 2, 0, 0,   /* (4) height = 600 */
		   1,  0,         /* (2) planes */
		   8,  0,         /* (2) colour depth */
		   0,  0, 0, 0,   /* (4) no compression */
		   0, 83, 7, 0,   /* (4) picture size, 800*600 */
		   0,  0, 0, 0,   /* (4) horizontal resolution */
		   0,  0, 0, 0,   /* (4) vertical resolution */		   		   
		   0,  0, 0, 0,   /* (4) colours used */
		   0,  0, 0, 0,   /* (4) important colours */
};				  /* total: 54 bytes */

/* BMP palette */

byte palette[1024];

long hdrlen, hdrused; /* hdrlen: Length of info area */
long pallen, palused, palbegin, palend;
long piclen, pich, picw;

static int isbmp(void)
{
	if ((hdr1[0] != 'B') || (hdr1[1] != 'M')) return 0;
	hdrlen = hdr1[14] + 256 * hdr1[15] + 65536 * hdr1[16] + 16777216 * hdr1[17];
	hdrused = (hdrlen < 40 ? hdrlen : 40);
	return 1;
}


static int is8bit(void)
{
	if ((hdr1[26] != 1) || (hdr1[28] != 8) || (hdr1[30] != 0)) return 0;
	pallen = hdr1[10] + 256 * hdr1[11] + 65536 * hdr1[12] + 16777216 * hdr1[13];
	pallen -= (hdrlen + 14);
	palused = (pallen < 1024 ? pallen : 1024);
	piclen = hdr1[2] + 256 * hdr1[3] + 65536 * hdr1[4] + 16777216 * hdr1[5];
	picw   = hdr1[18] + 256 * hdr1[19] + 65536 * hdr1[20] + 16777216 * hdr1[21];
	pich   = hdr1[22] + 256 * hdr1[23] + 65536 * hdr1[24] + 16777216 * hdr1[25];
	while (picw & 3) ++picw;	

	palbegin = hdrlen   + 14;
	palend   = palbegin + pallen;
	return 1;

}



static void get_hdr1(void)
{
	memcpy(&hdr1, &stdhdr, 54);
}



static void get_palette(void)
{
	int n, r, g, b;

	memset(palette, 0, sizeof(palette));
	for (n = 0; n < 256; n++)
	{
		GrQueryColor(n, &r, &g, &b);
		palette[n*4    ] = b;
		palette[n*4 + 1] = g;
		palette[n*4 + 2] = r;
	}
}


static byte colmap[256];	// Mappings of BMP colours to JOYCE colours



static void put_palette(void)
{
	int n, m;
	int r, g, b;

	for (n = 0; n < 256; n++)
	{
		b = palette[n*4    ];
		g = palette[n*4 + 1];
		r = palette[n*4 + 2];
		m = colour_find(r,g,b);
		colmap[n] = m;
	}
}


/* Generate any 128-byte record from a .BMP file */

word get_bmp_record(word Z80addr, word recno)
{
	int n;
	long where = recno * 128;

	if (where > (800*600 + sizeof(palette) + sizeof(hdr1))) return 0;

	get_hdr1();
	get_palette();

	for (n = 0; n < 128; n++)
	{
		register int m = n + where;

		if (m < sizeof(hdr1)) 
			WrZ80(Z80addr++, hdr1[m]);
		else if (m < (1024 + sizeof(hdr1)))
			WrZ80(Z80addr++, palette[m - sizeof(hdr1)]);
		else	
		{
			int px, py, pz;

			pz = (m - sizeof(hdr1) - sizeof(palette));
			py = (pz / 800);
			px = (pz % 800);

			WrZ80(Z80addr++, GrPixel(px,599 - py));
		}
	}
	return 1;
}




word put_bmp_record(word Z80addr, word recno)
{
	int bmpptr, z80ptr = 0;
	int where = 128 * recno;
	static int loadpal = 0;

	if (!recno)
	{
/*
 * Record 0: Load the header, which gives the sizes of the other objects in
 *          the BMP. This is rather complicated.
 * 
 * 1. Read in the header (14 bytes) and the first dword of the info area
 *   (4 bytes)
 *
 */
		for (z80ptr = 0; z80ptr < 18; z80ptr++)
		{
			hdr1[z80ptr] = RdZ80(Z80addr + z80ptr);
		}
		if (!isbmp()) return 0;
/*
 * hdrlen  = number of bytes in info area 
 * hdrused = number of info bytes we can use ( <= hdrlen)
 *
 * 2. Read in the rest of the info area ((hdrused - 4) bytes)
 * 
 */
		for (z80ptr = 18; z80ptr < 14 + hdrused; z80ptr++)
		{
			hdr1[z80ptr] = RdZ80(Z80addr + z80ptr);
		}
		if (!is8bit()) return 0;
		z80ptr = hdrused + 14;	/* header + used info area */
	}
/*
 * 3. Read palette or picture data
 */
	while (z80ptr < 128)
	{
		bmpptr = where + z80ptr;
		if      (bmpptr < hdrlen + 14) continue; /* Skipping bits of header */
		else if (bmpptr < palend) /* Loading palette */
		{
			if (bmpptr - palbegin < 1024)
			{
				palette[bmpptr - palbegin] = RdZ80(Z80addr + z80ptr);
			}
			loadpal = 1;
		}
		else if (bmpptr < palend) /* Skipping bits of palette */
		{	
			++z80ptr;
			continue;
		}
		else	/* Part of the picture - plot it */
		{
			int px, py, pz;

			if (loadpal)
			{
				put_palette();
				loadpal = 0;			
			}
			pz = (bmpptr - palend);
			py = (pz / picw);
			px = (pz % picw);

			GrPlot(px, pich - py - 1, colmap[RdZ80(Z80addr + z80ptr)]);
		}
		++z80ptr;
	}
	if (where + 128 > piclen) 
	{
	        SDL_UpdateRect(screen, 0, 0, 0, 0);
		return 0;
	}
	return 1;
}
