/** JOYCE PCW emulator for Linux / SDL ***********************/
/**                                                         **/
/** JOYCEMNU.C - handles the menu system and the debugging  **/
/**             console                                     **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-2000                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "Joycevga.h"
#include "Joyceset.h"
#include "keyboard.h"
#include "inline.h"


/* XBM files from xpaint(1) define the image as an array of "unsigned char", 
  which compiles without warnings. XBM files from xv(1) define it as 
  "char", which causes lots of warnings. */

#include "pcwkey.xbm"	/* X-Window bitmap, a graphic format designed to
			 * be used in C source.
			 */

/* Routines for JOYCE to draw on the 800x600 screen. Similar to those in
   JOYCEVGA, but for the menu system. */

static short vid_y=0, vid_x=0;
static text_rv=0; 

/* Set reverse video */

void vid_rv(char r)
{
	if (r && text_rv) return;
	if (!r && !text_rv) return;

	text_rv=text_fg;
	text_fg=text_bg;
	text_bg=text_rv;
	text_rv=r;
}

/* Draw character */

void vid_putch(char c)
{
	byte *sb;
	int ox = vid_x, oy = vid_y, ow = 8, oh = 16;

        if ((mustlock || context == screen) && SDL_LockSurface(context) < 0 ) return;

	if (c==1) vid_rv(0);
	else if (c==2) vid_rv(1);
	else if (c=='\n')
	{
		GrFilledBoxNC(vid_x, vid_y, 799, vid_y + 15, text_bg); /* (v1.20) */

		ox = 0; ow = 800;
 		vid_x=0;
		vid_y+=16;
		if (vid_y>584) vid_y=0;
	}
	else 
	{
		sb=scrfont+16*c;
		draw_bitmap(vid_y, vid_x, 16, 8, sb, text_fg, text_bg);
		vid_x+=8;
		if (vid_x>792)
		{
			vid_x=0;
			vid_y+=16;
			if (vid_y>584) vid_y=0;
		}
	}
	if (context == screen) 
	{
		if (mustlock) SDL_UnlockSurface(context);
		SDL_UpdateRect(context, ox, oy, ow, oh);
	}

}

/* Move cursor */

void vid_at(char y,char x)
{
	vid_y=y*16;
	vid_x=x*8;
}

/* Put string */

void vid_puts(char *s)
{
	while(*s) vid_putch(*s++);	
}

/* Put string at y,x */

void vid_putsxy(char y, char x, char *s)
{
	vid_at(y,x);
	vid_puts(s);
}




void vid_packsxy(char y, char x, char *s)	/* Print text, clear the */
{						/* rest of the line */
	vid_putsxy(y,x,s);

	if ((mustlock || context == screen) && SDL_LockSurface(context) < 0 ) return;
	GrFilledBoxNC((x + strlen(s)) * 8, y * 16, 799, y * 16 + 15, text_bg);	/* v1.20 */
        if (context == screen)
        {
                if (mustlock) SDL_UnlockSurface(context);
                SDL_UpdateRect(context, 0, y * 16, 800, 16);
        }
}



int dgetstr(int y, int x, char *S, int maxc)	/* Input string */
{						
	char c[2];
	int L = 0;

	c[1] = 0;
	while(1)
	{
		c[0] = dgetch();
		if (c[0] == '\n' || c[0] == '\r')
		{
			S[L++] = 0;
			return 1;
		}
		if (c[0] == 0x1B)
		{
			S[L++] = 0;
			return 0;
		}
		if (c[0] == 8 || c[0] == 127)
		{
			if (L > 0)
			{
				--L;
				--x;
				vid_putsxy(y,x," ");
			}
		}
		else
		{
			if (L < maxc - 1)
			{
				S[L++] = c[0];
				vid_putsxy(y,x,c);
				++x;
			}
		}
	}
}		
	


/* Back up screen display before drawing menus etc. on it */

void SaveBuf1(void)
{
	if (ScrType) SaveBuf();
}


static void KeyDraw(void)	/* Draw the keyboard map */
{
	int y,x,b,z=0;
	byte ch;

        if ((mustlock || context == screen) && SDL_LockSurface(context) < 0 ) return;

	for (y = 0; y < pcwkey_height; y++)
		for (x = 0; x < pcwkey_width; x+=8)
		{
			ch = pcwkey_bits[z++];
			for (b = 0; b < 8; b++)
			{
				GrPlotNC(x+b, y+32, (ch & 1) ? text_bg : text_fg);
				ch = (ch >> 1);
			}
		}
        if (context == screen)
        {
                if (mustlock) SDL_UnlockSurface(context);
                SDL_UpdateRect(context, 0, 32, pcwkey_width, pcwkey_height);
        }
}




/* Ask the user if they want to quit. Returns only if they don't */

int Confirm(char *s)
{
	volatile char c;

	vid_rv(0);
	vid_packsxy(0,0,s);
	vid_packsxy(1,0," ");
	while(1)
	{
		c = dgetch();
		if (c == 'Y' || c == 'y')
		{
			return 1;
		}
		if (c == 'N' || c == 'n')
		{
			return 0;
		}
	}
}





void ExitConf(void)
{
	static char mutex=0;

	if (mutex) return;

	mutex=1;
	SaveBuf1();
	if (Confirm("Press Y to quit JOYCE, N to return")) goodbye(99);
	BlitBuf();
        UpdateArea(0,0,0,0);
}

/* For dumping the CPU state */

static void putFlags(char *s, byte F)
{
	strcat(s, F & S_FLAG ? "S" : "s");
	strcat(s, F & Z_FLAG ? "Z" : "z");
	strcat(s, F & H_FLAG ? "H" : "h");
	strcat(s, F & P_FLAG ? "P" : "p");
	strcat(s, F & V_FLAG ? "V" : "v");
	strcat(s, F & N_FLAG ? "N" : "n");
	strcat(s, F & C_FLAG ? "C" : "c");
}

/* Dump the CPU state */

void CPUstate(Z80 *R)
{
	char s[104];
	byte c;
	word t[3];

	t[0] = RdZ80(R->SP.W  ) + (RdZ80(R->SP.W+1) << 8);
	t[1] = RdZ80(R->SP.W+2) + (RdZ80(R->SP.W+3) << 8);
	t[2] = RdZ80(R->SP.W+4) + (RdZ80(R->SP.W+5) << 8);

	sprintf(s,"\1A=%2.2x BC=%4.4x DE=%4.4x HL=%4.4x IX=%4.4x "
                  "IY=%4.4x PC=%4.4x SP=%4.4x -> %4.4x %4.4x %4.4x   ",
		  R->AF.B.h, R->BC.W, R->DE.W, R->HL.W, R->IX.W, R->IY.W, 
                  R->PC.W, R->SP.W, t[1], t[2], t[3]);
	putFlags(s, R->AF.B.l);
	strcat(s,"\n");
	vid_putsxy(0,0,s);
	sprintf(s,"\1A'%2.2x BC'%4.4x DE'%4.4x HL'%4.4x I=%2.2x "
                  "IFF=%2.2x  %2.2x %2.2x %2.2x %2.2x | %2.2x %2.2x "
                  "%2.2x %2.2x %2.2x %2.2x %2.2x     ", R->AF1.B.h, R->BC1.W, 
                  R->DE1.W, R->HL1.W, R->I, R->IFF, RdZ80(R->PC.W-4), 
                  RdZ80(R->PC.W-3), RdZ80(R->PC.W-2), RdZ80(R->PC.W-1), 
                  RdZ80(R->PC.W), RdZ80(R->PC.W+1), RdZ80(R->PC.W+2),
		  RdZ80(R->PC.W+3), RdZ80(R->PC.W+4), RdZ80(R->PC.W+5), 
                  RdZ80(R->PC.W+6) /*" "*/);

	putFlags(s, R->AF1.B.l);
	strcat(s, "    ESC=Exit");
	vid_putsxy(1,0,s);

	c=0;
	while (c != 27) c = dgetch();
}



void CenSettings(void)	/* CEN settings */
{
	char cenport[101];
	int refresh = 0;
	char *str;

	while(1)
	{

		vid_rv(0);
		if (!jset.cen_port & 0x80)
		{
			jset.cen_port = 0x80;
			sprintf(jset.printer_file, "/dev/null");
		}
		sprintf(cenport,"CEN is mapped to file %-77.77s",
					jset.printer_file);
		vid_packsxy(0,0,cenport);
		vid_putsxy(1,0,"\1"
			       "F=Choose file                                     "
			       "                                          ESC=Exit");
		refresh = 0;
		while(!refresh)
		{
			char c = dgetch();
			switch(c)
			{
				case 'F': case 'f':
				vid_packsxy(0,0,"Enter path for file");
				vid_packsxy(1,0,"Press ESC to cancel");
//				vid_rv(1);
//				vid_packsxy(0,22," ");
				str = Itemsel("*.*"); /* v1.20 */			
				if (str)
				{
					strcpy(jset.printer_file, str);
					jset.cen_port |= 0x80;
				}
				refresh = 1;
				break;
			
				case 0x1B:
				PtrSettings();
				return;
			}
		}
	}
}


void ComSettings(void)	/* (v1.31) Com settings */
{
	char comport[101];
	char *str;
	int refresh = 0;

	while(1)
	{

		vid_rv(0);      
		sprintf(comport,"Output to %-40.40s    Input %-40.40s", jset.com_out, jset.com_in);
		vid_packsxy(0,0,comport);
		vid_putsxy(1,0,"\1"
			       "O=Output file  I=Input file                       "
			       "                                          ESC=Exit");
		refresh = 0;
		while(!refresh)
		{
			char c = dgetch();
			switch(c)
			{
				case 'I': case 'i':
                                vid_packsxy(0,0,"Enter path for file");
                                vid_packsxy(1,0,"Press ESC to cancel");
                                str = Itemsel("*"); /* v1.20 */
                                if (str)
                                {
                                        strcpy(jset.com_in, str);
                                }
                                refresh = 1;
                                break;

                                case 'O': case 'o':
                                vid_packsxy(0,0,"Enter path for file");
                                vid_packsxy(1,0,"Press ESC to cancel");
                                str = Itemsel("*"); /* v1.20 */
                                if (str)
                                {
                                        strcpy(jset.com_out, str);
                                }
                                refresh = 1;
                                break;

				case 0x1B:
				set_com();
				return;
			}
		}
	}
}




void KbdSettings(void)	/* Keyboard settings */
{
	int refresh = 0;

	while(1)
	{

		vid_rv(0);
		if (jset.swap_alt) vid_putsxy(0,0, "Host CTRL -> PCW ALT    Host ALT -> PCW EXTRA     ");
		else	           vid_putsxy(0,0, "Host CTRL -> PCW EXTRA  Host ALT -> PCW ALT       ");
                if (jset.swap_delete) 
			vid_packsxy(0, 50, "Backspace and Delete swapped.");
                else 	vid_packsxy(0, 50, "Backspace and Delete not swapped.");

		vid_putsxy(1,0,"\1"
			       "TAB=swap CTRL and ALT  D=swap Backspace and Delete"
			       "                                          ESC=Exit");
		refresh = 0;
		while(!refresh)
		{
			char c = dgetch();
			switch(c)
			{
				case 9:
				jset.swap_alt = !(jset.swap_alt);
				refresh = 1;
				break;

				case 'D': case 'd':
				jset.swap_delete = !(jset.swap_delete);
				refresh = 1;
				break;
			
				case 0x1B:
				KeySet();	/* Write new settings to keyboard */
				return;
			}
		}
	}
}



void BeepSettings(void)	/* v1.20: Beeper settings */
{
	int refresh = 0;

	while(1)
	{

		vid_rv(0);
		if (jset.beep)     vid_packsxy(0,0,"No beeper");
		else	           vid_packsxy(0,0,"Beeper flashes the screen");

		vid_putsxy(1,0,"\1"
			       "TAB to change beeper operation                    "
			       "                                          ESC=Exit");
		refresh = 0;
		while(!refresh)
		{
			char c = dgetch();
			switch(c)
			{
				case 9:
				jset.beep = 1 - jset.beep;
				refresh = 1;
				break;
			
				case 0x1B:
				return;
			}
		}
	}
}










void MouSettings(void)	/* Mouse settings */
{
	int refresh = 0;
	char cbuf[60];

	while(1)
	{

		vid_rv(0);				 
		if (jset.mousemode < 0)   vid_putsxy(0,0,"Mouse disabled");
		else if (!jset.mousemode) vid_putsxy(0,0,"Kempston mouse");
		else			  vid_putsxy(0,0,"AMX mouse     ");		

		if (MouseButtons < 0)       vid_packsxy(0,16,"No host mouse");
		else if (MouseButtons == 3) vid_packsxy(0,16,"Host mouse has 3 buttons - AMX mode recommended");
		else                        vid_packsxy(0,16,"Host mouse has 2 buttons - Kempston mode recommended");

		vid_putsxy(1,0,"\1"
			       "TAB to change mouse mode                          "
			       "                                          ESC=Exit");

		/* (v1.31) avoid divide-by-0 errors */
		if (jset.mxscale == 0) jset.mxscale = 8;
		if (jset.myscale == 0) jset.myscale = 8;

		/* (v1.23) extra mouse dynamics options for AMX */
		sprintf(cbuf,"X speed %d  Y speed %d  Warp speed %d", 
                        256 / jset.mxscale, 256 / jset.myscale, jset.mxspeed);
		if (jset.mousemode == 1) vid_putsxy(1,27,cbuf);

		/* End AMX extras */

		/* (v1.27) Mouse auto-patch */
		sprintf(cbuf, " A = Auto-patch (%s)", jset.patchmouse ? "on" : "off");
		vid_putsxy(0, 79 + jset.patchmouse, cbuf);
		/* End auto-patch */

		refresh = 0;
		while(!refresh)
		{
			char c = dgetch();
			switch(c)
			{
				case 9:
				++jset.mousemode;
				if (jset.mousemode == 2) jset.mousemode = (-1);
				refresh = 1;
				break;
			
				case 'X': case 'x':
				if (jset.mousemode != 1) break;
				--jset.mxscale; if (!jset.mxscale) jset.mxscale = 15;
				refresh = 1;
				break;

				case 'Y': case 'y':
				if (jset.mousemode != 1) break;
				--jset.myscale; if (!jset.myscale) jset.myscale = 15;
				refresh = 1;
				break;

				case 'W': case 'w':
				if (jset.mousemode != 1) break;
				++jset.mxspeed; 
				if (jset.mxspeed > 15) jset.mxspeed = 1;
				jset.myspeed = jset.mxspeed;
				refresh = 1;
				break;

				case 'A': case 'a':
				jset.patchmouse = ((!jset.patchmouse) & 1);
				refresh = 1;
				break;

				case 0x1B:
				return;
			}
		}
	}
}




void DskSettings(void)	/* Disk settings */
{
	int refresh = 0;
	char txt[101];

	while(1)
	{

		vid_rv(0);
		sprintf(txt,"Start up with %s  Boot discs are %s",
			    (jset.one_drive ? "one drive " : "two drives"),
			    (jset.boot_ro   ? "read-only " : "read/write"));
		vid_packsxy(0,0,txt);
		vid_putsxy(1,0,"\1"
			       "1, 2 set number of drives at startup   W=Write Pro"
			       "tect                                      ESC=Exit");
		refresh = 0;
		while(!refresh)
		{
			char c = dgetch();
			switch(c)
			{
				case '1': 
				jset.one_drive = 1;
				refresh = 1;
				break;

				case '2': 
				jset.one_drive = 0;
				refresh = 1;
				break;

				case 'W': case 'w': 
				jset.boot_ro = !(jset.boot_ro);
				refresh = 1;
				break;

				case 0x1B:
				return;
			}
		}
	}
}









void DskImage(void)	/* Insert/remove disc image file */
{
	int refresh;
	static byte cdrv = 0;
	static char ImgPath[4][79];
	char txt[101];

	while(1)
		{
		vid_rv(0);
		vid_putsxy(1,0,"A=Select drive A:   B=Select drive B:   E=Eject   "
			       "I=Insert  W=Write Protect  D=Disable      ESC=Exit");
		
		sprintf(txt,"Drive %c:  %s  %s", (cdrv + 'A'),
				(drive_readonly[cdrv] ? "Read Only " : "Read/Write"),	
				(drive_filename[cdrv] ? drive_filename[cdrv] :
					(drive_readonly[cdrv] ? "[Ejected]" : "[Disabled]")
				)
			);
		vid_packsxy(0,0,txt);

		/* 'Insert' disabled if disc in drive */
		if (drive_filename[cdrv]) vid_putsxy(1,50,"\1        ");	
		refresh = 0;
		while(!refresh)
		{
			char c = dgetch();
			switch(c)
			{
				char *str;

				case 'A': case 'a':
				case 'B': case 'b':
				cdrv = ((c - 'A') & 1);
				refresh = 1;
				break;

				case 'W': case 'w':
				if (drive_filename[cdrv]) 
				{
					drive_readonly[cdrv] ^= 1;
					refresh = 1;
				}				
				break;

				case 'E': case 'e':
				if (drive_filename[cdrv])
				img_eject(cdrv);
				drive_readonly[cdrv] = 1;
				refresh = 1;
				break;

				case 'D': case 'd':
				img_eject(cdrv);
				drive_readonly[cdrv] = 0;
				refresh = 1;
				break;

				case 'I': case 'i':
				if (drive_filename[cdrv]) break;
				vid_packsxy(0,0,"Enter path for image");
				vid_packsxy(1,0,"Press ESC to cancel");
				//vid_rv(1);
				//vid_packsxy(0,22," ");
				str = Itemsel("*.DSK"); /* v1.20 */
				if (str)
				{
					FILE *dfp;	
					strcpy(ImgPath[cdrv], str);	
					drive_filename[cdrv] = ImgPath[cdrv];
					drive_readonly[cdrv] = 0;
		
					dfp = fopen(drive_filename[cdrv], "rb");
					if (!dfp)
					{	
						char c;
						vid_rv(0);
						vid_packsxy(1,0,"This file does not exist. Create it (Y/N)?");
						do {
							c = dgetch();
							if (c == 'N' || c == 'n')
							{
								img_eject(cdrv);
								drive_readonly[cdrv] = 1;
							}
							if (c == 'Y' || c == 'y') 
								if (!img_new(cdrv))
								{
									vid_packsxy(1, 0, "Failed to create file. Press SPACE");
									dgetch();
									img_eject(cdrv);
									drive_readonly[cdrv] = 1;
									c = 'N';
								}
						} while (c != 'Y' && c != 'y' && c != 'n' && c != 'N');
					}
					else fclose(dfp);
				}
				vid_rv(0);
				refresh = 1;
				break;

				case 0x1B:
				return;
			}
		}
	}


}




void Settings(void)	/* The settings menu */
{

	int refresh;
	while(1)
		{
		vid_putsxy(0,0,"\1"
			       "JOYCE settings                                    "
		       	       "                                                  ");
		vid_putsxy(1,0,"\1"
			       "C=CEN port  K=Keyboard  D=Disks   M=Mouse   B=Beep"
			       "er  O=cOm port                    S=Save  ESC=Exit");

		refresh = 0;
		while(!refresh)
		{
			char c = dgetch();
			switch(c)
			{
				case 'C': case 'c':	
				CenSettings();
				refresh = 1;
				break;
			
				case 'K': case 'k':
				KbdSettings();
				refresh = 1;
				break;

				case 'D': case 'd':
				DskSettings();
				refresh = 1;
				break;

				case 'O': case 'o': /* v1.21 */
				ComSettings();
				refresh = 1;
				break;

				case 'M': case 'm':
				MouSettings();
				refresh = 1;
				break;

				case 'B': case 'b':
				BeepSettings();
				refresh = 1;
				break;

				case 'S': case 's':
				if (write_settings())
					vid_putsxy(0,0,"\1Settings were saved    ");
				else	vid_putsxy(0,0,"\1Failed to save settings");
				break;

				case 0x1B:
				return;
			}
		}
	}
}


/* Main menu */

void MenuDraw(void)	/* Draw the menu */
{
	char menustr[100];

	sprintf(menustr,"JOYCE v%x.%02x   Copyright 1996-2000, John Elliott  "
                        "Compiled with SDL version %d.%d.%d",
			BCDVERS >> 8,
			BCDVERS & 0xFF, 
			SDL_MAJOR_VERSION, 
			SDL_MINOR_VERSION,
			SDL_PATCHLEVEL );
	vid_rv(0);
	vid_packsxy(0,0,menustr);
	vid_packsxy(1,0, "ESC=Return to JOYCE  I=dIsks  S=Settings  B=Boot  "
                         "C=CPU state  f10=Quit  K=Keyboard map");
#ifdef DEBUG				       
	vid_putsxy(1,89,"D=Debug");
#endif
}

void JoyceMenu(Z80 *R)	/* Handle the menu */
{
	static char mutex = 0;
	volatile byte c;

	if (mutex) return;	/* Prevent recursion */

	mutex=1;
	SaveBuf1();
	MenuDraw();
	c=0;
	while (!c)
	{
		c=dgetch();
		switch(c)
		{
			case 'D': case 'd':	/* D = Debug mode */
			DebugMode = !DebugMode;
			if(DebugMode) R->Trace = 1;
			shadow_invalid = 1;	/* Force screen repaint */
			Border();
			dcls();			/* FALL THROUGH */
			
			case 27:		/* ESC */
			BlitBuf();
			UpdateArea(0,0,0,0);
			mutex=0;
			return;
			break;

			case 'B': case 'b':	/* B = Boot */
			if (Confirm("Press Y to reboot the PCW, N to cancel"))
			{
				BlitBuf();
	                        UpdateArea(0,0,0,0);
				mutex=0;
				jstate.reset_flag=1;
				return;
			}
			MenuDraw();
			c=0;
			break;

			case 'I': case 'i':	/* I = disk Images menu */
			DskImage();
			MenuDraw();
			c=0;
			break;

			case 'K': case 'k': 	/* K = show keyboard */
			KeyDraw();
			c=0;
			break;

			case 'S': case 's':	/* S = Settings */
			Settings();
			MenuDraw();
			c=0;
			break;		

			case 'C': case 'c':	/* C = CPU state */
			CPUstate(R);
			MenuDraw();
			c=0;
			break;

			case 10:	/* f10 = Quit */
			if (Confirm("Press Y to quit JOYCE, N to return")) goodbye(99);
			MenuDraw();

			/* FALL THROUGH to */

			default:
			c=0;
			break;
		}
	}
}

char message[2048];		/* Used in joyce_dprintf*() */

static int dy = 19, dx = 0;	/* cursor coords for debugger console */

static void dputch(char c)	/* putch() for the debugger console */
{
	int a, b, x, y;
	int ox, oy, ow = 8, oh = 16;

	if (framebuf == NULL) return;

	oy = y = dy * 16;
	ox = x = dx * 8;
	context = framebuf;

        if ((mustlock || context == screen) && SDL_LockSurface(context) < 0 ) return;

	for (a = 0; a < 16; a++)
	{
		byte v = scrfont[c*16+a];
		for (b = 0; b < 8; b++)
		{
			GrPlotNC(x+b, y+a, (v & 0x80 ? debug_fg : debug_bg));
			v = (v << 1);
		}
	}
        if (context == screen)
        {
                if (mustlock) SDL_UnlockSurface(context);
                SDL_UpdateRect(context, ox, oy, ow, oh);
        }
	context = screen;
}


static void dputs(char *str)	/* puts() for the debugger console */
{
	int p = 0;
	while (str[p])
	{
		if (str[p] == '\n') { dx = 0; ++dy; }
		else if (str[p] == 8) { if (dx > 0) --dx; }
		else	
		{
			dputch(str[p]);		
			dx++;
			if (dx > 100) { dx = 0; ++dy; }
		}
		if (dy > 36)
		{
			SDL_Rect sr, dr;

			sr.x = dr.x = 0;
			dr.y = 300; 
                        sr.y = 316;
			sr.w = dr.w = 800;
			sr.h = dr.h = 284;

			if (framebuf == NULL) return;

/*		        SDL_MapSurface(framebuf, framebuf->format); */
			SDL_BlitSurface(framebuf, &sr, framebuf, &dr);
/*			SDL_MapSurface(framebuf, screen->format); */

                        if (mustlock && SDL_LockSurface(framebuf) < 0 ) return;
			context = framebuf;
			GrFilledBoxNC(0, 576, 799, 599, colour_find(0,0,0));
			context = screen;
		        if (mustlock) SDL_UnlockSurface(framebuf);
			dy = 36;
		}	
		++p;
	}	
}

void joyce_dprintf(char *s, ...)	/* printf() for the debugger console */
{
	va_list arg;

	va_start(arg, s);
	if (!DebugMode || !ScInited)	/* If debugger not active, or screen
					  not set up, write to log file */
	{
		if (fplog) 
		{
			vfprintf(fplog, s, arg);
			fflush(fplog);
		}
	}
	else
	{
		vsprintf(message, s, arg);
		dputs(message);
	}

	va_end(arg);
}

void dgets(char *S, int maxc)	/* Gets() for the debugger console */
{
	int L = 0;

	char c[2];

	c[1] = 0;
	while(1)
	{
		c[0] = dgetch();
		if (c[0] == '\n' || c[0] == '\r')
		{
			S[L++] = 0;
			joyce_dprintf("\n");
			DrawScr();
			return;
		}
		if (c[0] == 8 || c[0] == 127)
		{
			if (L > 0)
			{
				--L;
				joyce_dprintf("\010 \010");
				DrawScr();
			}
		}
		else
		{
			joyce_dprintf(c);
			DrawScr();
			if (L < maxc - 1) S[L++] = c[0];
		}
	}
}		


char dgetch(void)
{
	unsigned char c = get_keytoken();

	switch(c)	/* (v1.20) Map some keys to WordStar control codes */
	{
		case SDLK_F10:		/* f10   */ c = 10; break;
		case SDLK_UP:		/* up    */ c =  5; break; 
		case SDLK_DOWN:		/* down  */ c = 24; break;
		case SDLK_LEFT:		/* left  */ c = 19; break;
		case SDLK_RIGHT:	/* right */ c =  4; break;
		case SDLK_DELETE:	/* del   */ c =  7; break;
		case SDLK_INSERT:	/* ins   */ c = 22; break;
		case SDLK_PAGEUP:	/* pgup  */ c = 18; break;
		case SDLK_PAGEDOWN:	/* pgdn  */ c =  3; break;
	}

	memset((byte *)keyboard_map, 0, sizeof(keyboard_map));
	keyboard_chain(0);
	return c;
}


void dcls(void)
{
	if (!DebugMode) return;

        if (mustlock && SDL_LockSurface(framebuf) < 0 ) return;

	context = framebuf;
	GrFilledBoxNC(0, 300, 799, 599, colour_find(0,0,0));
	context = screen;

	if (mustlock) SDL_UnlockSurface(framebuf);
}


/****************** (v1.27) Support for popup messages ****************/

void DrawPopup(void)
{
	char pbuff[81];
	int l, x, w, y, ty;
	time_t t;

	l = strlen(jstate.popup);
	if (l > 80) l = 80;
	sprintf(pbuff, "%-*.*s", l, l, jstate.popup);

	x = (400 - 4*l) - 8;
	w = 8*l + 16;

	if (jstate.popup_at_bottom) 
	{
		y  = 556; 
		ty = 35;
	}
	else
	{
		y  = 300;
		ty = 19;
	}

        if ((mustlock || context == screen) && SDL_LockSurface(context) < 0 ) return;

	GrFilledBoxNC(x,     y,     x + w,     y + 24, text_bg);

	GrBoxNC      (x,     y,     x + w,     y + 24, text_fg);
        Border3d     (x - 1, y - 1, x + w + 1, y + 25, 1);
        Border3d     (x - 2, y - 2, x + w + 2, y + 26, 1);
        Border3d     (x - 3, y - 3, x + w + 3, y + 27, 1);

	vid_putsxy(ty, 50 - (l/2), pbuff);
        if (context == screen)
        {
                if (mustlock) SDL_UnlockSurface(context);
                SDL_UpdateRect(context, x, y, w, 24);
        }

	time(&t);
	if (t == jstate.popup_timeout) popup_off();
}


void popup_on(char *msg, int secs, int centred)
{
	time_t t;

	if (ScrType) 
	{
		joyce_dprintf(msg);
		return;
	}

	t = 0;
	if (secs) 
	{
		time(&t);
		t += secs;
	}
	strncpy(jstate.popup, msg, 80);
	jstate.popup_timeout = t;
	jstate.popup_at_bottom = !centred;
}

void popup_off(void)
{
	shadow_invalid = 1;
	if (jstate.popup_at_bottom) Border();
	jstate.popup_timeout = 0;
	jstate.popup_at_bottom = 0;
	jstate.popup[0] = 0;
}
