/** JOYCE PCW emulator for Linux/SDL *************************/
/**                                                         **/
/** JoyceSdlScr.c: contains the screen drawing code for SDL **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1999                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#define SCRMOD
#include "Joyce.h"
#include "Joycevga.h"

#include "mask.xbm"

void flip_xbm(int w, int h, byte * bits);


/* Palette reloading */
/* 
 * Note: There is a bug in SDL 0.10.0 which makes it impossible to change
 *      palettes. This is nasty as it means that XPALETTE will probably
 *      try to select colours >64, SDL will fail silently and keep these
 *      colours as black, and the JOYCE screen will go blank!
 */

#if (SDL_MAJOR_VERSION == 0) && (SDL_MINOR_VERSION == 10) && (SDL_PATCHLEVEL == 0)
#error SDL 0.10.0 is buggy. JOYCE will not work with this version of SDL.
#endif

void reload_palette(void)
{
        SDL_SetColors(framebuf, colours, 0, 256);
/*	SDL_MapSurface(framebuf, screen->format); */
}


static SDL_Surface *iconsurface;


char InitScr(int type)	/* Initialise the screen */
{
	SDL_Palette *palette;
	int nr, ng, nb;
        static char title[50];

        ScrEnabled = 1;

        if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) 
	{
                joyce_dprintf("Couldn't initialize SDL: %s\n", SDL_GetError());
		return 0;
        }

        /* Have a preference for 8-bit, but accept any depth */
	if (screen == NULL)
	{
	        screen = SDL_SetVideoMode(800, 600, 8, 
                                            SDL_SWSURFACE | SDL_ANYFORMAT);
	}
        if ( screen == NULL ) 
	{
                joyce_dprintf("Couldn't set 800x600x8 video mode: %s\n",
                                                        SDL_GetError());
		return 0;
        }
        joyce_dprintf("Set 800x600 at %d bits-per-pixel mode\n",
                                        screen->format->BitsPerPixel);

	if (framebuf == NULL)
	{
		mustlock = 0;
		if (screen->format->BitsPerPixel == 8)
		{
			framebuf = SDL_AllocSurface(SDL_HWSURFACE, 800, 600, 8, 
                           0, 0, 0, 0);
			if (framebuf == NULL) return 0;
			if (framebuf->flags & SDL_HWSURFACE) mustlock = 1;
		}
		else 
		{
			framebuf = SDL_AllocSurface(SDL_SWSURFACE, 800, 600, 8,
                           0, 0, 0, 0);
			if (framebuf == NULL) return 0;
		}
	}
	context = screen;

	linelen = framebuf->pitch;

/* Init colours */

       /* Get the current colormap */
        palette = framebuf->format->palette;
        if ( palette == NULL ) 
	{        /* The frame buffer is not palettized; which is pretty *
                  * weird, considering  we asked for 8-bit palettized. */
                return 0;
        }
        ncolours = palette->ncolors;
        colours = (SDL_Color *)malloc(ncolours*sizeof(SDL_Color));
        memcpy(colours, palette->colors, ncolours*sizeof(SDL_Color));

        for (nr = 0 ; nr < 256; nr += 85)
                for (ng = 0; ng < 256; ng += 85)
                        for (nb = 0; nb < 256; nb += 85)
                                colour_find(nr, ng, nb);

/* Got the original palette */

        colour_nextfree= 0;
        ScrType        = type;
        rram_base      = PCWRAM;
        scr_h          = 256;
        y_origin       = 0;
        colour_reg     = 0;
        dblk = blk     = colour_find(0, 0, 0);
        dwhi = whi     = colour_find(0xFF, 0xFF, 0xFF);
        colour_nextfree=65;
	reload_palette();

        Border();
        GrClearScreen(blk);
        shadow_invalid = 1;     /* (v1.30) */

/* Set up the colours of JOYCE screen elements, using the "guaranteed" 
  palette */

        border_beep = colour_find(255,   0,   0);
        text_fg     = colour_find(170, 255, 255);
        text_bg     = colour_find(  0,   0, 170);
        debug_fg    = colour_find(  0, 255,   0);
        debug_bg    = colour_find(  0,   0,   0);
        shadow_hi   = colour_find( 85,  85, 170);
        shadow_lo   = colour_find(  0,   0,  85);
	reload_palette();

	if (!iconsurface) 
	{
		flip_xbm(mask_width, mask_height, mask_bits);
		iconsurface = SDL_LoadBMP(LIBROOT "/pcw.bmp");
	}
	if (iconsurface) SDL_WM_SetIcon(iconsurface, mask_bits);
	sprintf(title, "JOYCE v%x.%x", BCDVERS >> 8, BCDVERS & 0xFF);
	SDL_WM_SetCaption(title, "Joyce");

        ScInited = 1;
        if(DebugMode) dcls();
        return 1;
}


void UnitScr(void)
{
        if (ScInited)
        {
		SDL_FreeSurface(framebuf);
		if (iconsurface) SDL_FreeSurface(iconsurface);
		SDL_Quit();
		free(colours);
        }
        ScInited=0;
}


void beep_on(void)
{
	beeperstat = 1;
	if (!jset.beep)		/* v1.20: Using jset.beep==0 not BeepType==1 */ 
	{
		Border();
		DrawScr();
	}
	else			/* v1.20 */
	{
/*		sound(2000);*/	/* Sort of approximately the right pitch, */
	}			/* but I was _not_ going to sit there all */
}				/* evening with two beepers going full */
				/* blast trying to get it right */


void beep_off(void)
{
	beeperstat = 0;
	if (!jset.beep) 	/* v1.20 */
	{
		Border();
		DrawScr();
	}
	else			/* v1.20 */
	{
/*		nosound();*/	/* v1.20 */
	}
}

void screen_on(void)
{
	ScrEnabled = 1;
	shadow_invalid = 1;	/* (v1.30) */
	DrawScr();
}

void screen_off(void)
{
	ScrEnabled = 0;
	DrawScr();
}

void screen_colour(int wr, int wg, int wb, int br, int bg, int bb)
{
	set_ink(&dblk, br, bg, bb);
	set_ink(&dwhi, wr, wg, wb);	
	shadow_invalid = 1;
}


void Border(void)	/* Draw the border in the frame buffer */
{
	int lblk;

	if (ScrType || !ScrEnabled) return;
	if (colour_reg & 0x80)	/* Colours inverted */
	{
		lblk = blk = dwhi;
		whi  = dblk;
	}
	else
	{
		lblk = blk = dblk;
		whi  = dwhi;
	}

	if (!jset.beep && beeperstat) /* v1.20 */
	{
		lblk = border_beep; 
	}

	context = framebuf;
	if (mustlock && SDL_LockSurface(context) < 0 )
	{
		context = screen; 
		return;
	}

	if (colour_reg & 0x40)	/* Border covers screen? */
	{
		if(DebugMode)
		{	
			GrFilledBoxNC(  0,   0, 799,  21, lblk);
			GrFilledBoxNC(  0,  22,  39, 277, lblk);
			GrFilledBoxNC(760,  22, 799, 277, lblk);
			GrFilledBoxNC(  0, 278, 799, 299, lblk);
		}
		else
		{
			GrFilledBoxNC(  0,   0, 799,  43, lblk);
			GrFilledBoxNC(  0,  43,  39, 556, lblk);
			GrFilledBoxNC(760,  43, 799, 556, lblk);
			GrFilledBoxNC(  0, 556, 799, 599, lblk);
		}
	}
	else
	{
		GrFilledBoxNC(0, 0, 800, (DebugMode ? 299 : 599),lblk);
	}
	if (mustlock) SDL_UnlockSurface(context);
	context = screen;
}
	



void BlitBuf(void)		/* Send the Framebuf to video RAM  */
{
	SDL_BlitSurface(framebuf, NULL, screen, NULL);
}


void SaveBuf(void)		/* Save the video RAM to Framebuf */
{
/*        SDL_MapSurface(screen, framebuf->format); */
        SDL_BlitSurface(screen, NULL, framebuf, NULL);
/*        SDL_MapSurface(framebuf, screen->format); */

}



static void DrawLine(unsigned char *bytes, short yoff) /* Draw a PCW line */
{
	int x;
	register int b;
	register byte v;
	long yp;
	byte *data = (byte *)framebuf->pixels;

	if (DebugMode)
	{
		yp = ((yoff+22)*linelen + 40);
		for (x=0;x<90;x++)
		{
			v=(*bytes);
			if (v != screen_shadow[x][yoff] || shadow_invalid) 
								/* (v1.30) */
			{
				screen_shadow[x][yoff] = v; /* (v1.30) */
				for (b=0;b<8;b++)
				{
					data[yp++]=(v & 0x80 ? whi:blk);
					v=(v<<1);
				}
			}
			else yp += 8;
			bytes+=8;
		}

	}
	else
	{
		yp = ((yoff*2+44)*linelen+40);
		for (x=0;x<90;x++)
		{
			v=(*bytes);
			if (v != screen_shadow[x][yoff] || shadow_invalid) /* (v1.30) */
			{
				screen_shadow[x][yoff] = v; /* (v1.30) */
				for (b=0;b<8;b++)
				{
					data[yp++]=(v & 0x80 ? whi:blk);
					v=(v<<1);
				}
			}
			else yp += 8;
			bytes+=8;
		}
		memcpy(data+yp+linelen-720, 
                       data+yp-720, linelen);
	}
}



void DrawScr(void)	/* Draw the PCW screen */
{
	short y;
	register int rrvalue;

	if (ScrType) return;
       
	if (mustlock && SDL_LockSurface(framebuf) < 0 ) return;

/* If you get problems with screen updates uncomment this line
 *
 *	shadow_invalid = 1;
 */
	if (!ScrEnabled)
	{
		memset(framebuf->pixels, dblk, 600*linelen);
		shadow_invalid = 1;	/* (v1.30) */
	}
	else if (colour_reg & 0x40)
	{
		context  = framebuf;
		for (y=0;y<scr_h;y++)
		{
			register short yy=((y+y_origin)&0xFF)<<1;

			rrvalue=rram_base[yy+1];
			rrvalue=rrvalue <<8;
			rrvalue+=rram_base[yy];
			DrawLine(PCWRAM+(rrvalue    << 1) - (rrvalue & 7),y);
		}
		context = screen;
		shadow_invalid = 0;
	}
	if (jstate.popup[0]) 
	{
		context = framebuf;
		DrawPopup();
		context = screen;
	}
	SDL_UnlockSurface(framebuf);
	BlitBuf();
	SDL_UpdateRect(screen, 0, 0, 0, 0);
}



void VidPort(unsigned char port, unsigned char value) /* I/O to a video port */
{
	static int prev_colour_reg = -1;	/* (v1.30) */

	switch(port)
	{
		case 0xF5:
		rram_base=&PCWRAM[(long)value*512L];
		break;

		case 0xF6:	/* Move Y origin */
		y_origin=value;
		break;

		case 0xF7:	/* Set colour register */
		colour_reg=value;
		if (prev_colour_reg != value)
		{
			shadow_invalid = 1;	/* (v1.30) */
			prev_colour_reg = value;
		}			
		Border();
		break;
	}
}

/************************************************************************/
/* Palette management */

int colour_find(int r, int g, int b)    /* Search for a colour - if not */
{                                       /* found, allocate it */
        int n;
        for (n = 0; n < 255; n++)
        {
                if (colours[n].r == r && colours[n].g == g && 
                    colours[n].b == b) return n;
        }
	colours[colour_nextfree].r = r;
	colours[colour_nextfree].g = g;
	colours[colour_nextfree].b = b;

/* The top 128 colours can be freed when we need to */

	n = colour_nextfree;
        ++colour_nextfree;
        if (colour_nextfree == 256) colour_nextfree = 128;
	return n;
}


void GrQueryColor(int n, int *r, int *g, int *b)
{
	*r = colours[n].r;
	*g = colours[n].g;
	*b = colours[n].b;
}


void set_ink(int * ink, int r, int g, int b)    /* Set foreground or background */
{
        int ocol = *ink;

        (*ink) = colour_find(r,g,b);

        if (whi == ocol) whi = (*ink);
        if (blk == ocol) blk = (*ink);

	reload_palette();
}


void flip_xbm(int w, int h, byte * bits)
{
	int n, m;
	byte b, b2;

	for (n = (((w + 7) / 8) * h) - 1; n >= 0; n--)
	{
		b = bits[n];
		b2 = 0;
		for (m = 0; m < 8; m++)
		{
                	b2 = b2 >> 1;
			if (b & 0x80) b2 |= 0x80;
			b  = b  << 1;
		}
		bits[n] = b2;
	}
}
