/** JOYCE PCW emulator for UNIX  *****************************/
/**                                                         **/
/** JoycecomUnix.c: Allow JOYCE access to a pair of files   **/
/** representing input & output channels of SIO             **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1999                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "inline.h"
#include <termios.h>
#include <sys/time.h>

static int fd_comin, fd_comout;

void com_ems(Z80 *R);	/* in JOYCEFID.C */

static int baud[] = { B50,  B75,   B110,  B134,  B150,  B300,
            	      B600, B1200, B1800, B2400, B2400, B4800,
                      B4800,B9600, B19200,B9600, B19200,B19200,
                      B19200,B57600,B115200};

static int spd = 14;	/* 9600 */

void setspeed(int speed)
{
	struct termios t;

	tcgetattr(fd_comin, &t);
	cfsetispeed(&t, baud[speed]);
	tcsetattr(fd_comin, TCSADRAIN, &t);

	tcgetattr(fd_comout, &t);
	cfsetospeed(&t, baud[speed]);
	tcsetattr(fd_comout, TCSADRAIN, &t);

	spd = speed;
}

static char input_stat(void)
{
        int i;
        fd_set rfds;
        struct timeval tv;

        FD_ZERO(&rfds);
        FD_SET(fd_comin, &rfds);

        tv.tv_sec = tv.tv_usec = 0;

        i = select(1, &rfds, NULL, NULL, &tv);

        if (i) return 1;        /* Data ready */
        else return 0;          /* Data not ready */
}

static char output_stat(void)
{
        int i;
        fd_set rfds;
        struct timeval tv;

        FD_ZERO(&rfds);
        FD_SET(fd_comout, &rfds);

        tv.tv_sec = tv.tv_usec = 0;

        i = select(1, NULL, &rfds, NULL, &tv);

        if (i) return 1;        /* Data ready */
        else return 0;          /* Data not ready */
}

static char com_input(void)
{
	char c;

	if (read(fd_comin, &c, 1) < 1) c = 0x1A;

	return c;	
}

static void com_output(char c)
{
	write(fd_comout, &c, 1);
}



void fid_com(Z80 *R)
{
	if (comstat)
	{
		switch(R->BC.B.l)
		{
			case 7: com_ems(R); break;
			case 8: break;		 /* COM port is a file*/
			case 9: R->AF.B.h = 0x80; break;
		}
		return;
	}

	switch(R->BC.B.l)
	{
		case 0: tcdrain(fd_comout); break;	/* Empty buffers */
		case 1: break;  /* Handshake */
		case 2: setspeed(R->HL.W); break;  /* Set baud */
		case 3: R->AF.B.h = input_stat();     break;
		case 4: R->AF.B.h = com_input();      break;
		case 5: R->AF.B.h = output_stat();    break;
		case 6: com_output(R->DE.B.l);	      break;
		case 7: com_ems(R);		      break;
		case 8: set_com();                    break;
		case 9: R->AF.B.h = 0x80;             break;

	}
}

void com_deinit()
{
	if (fd_comin  >= 0) close(fd_comin);
	if (fd_comout >= 0) close(fd_comout);
}

void com_init()
{
	struct termios t;

	if (jset.com_in[0]  == 0) strcpy(jset.com_in,   "/dev/zero");
	if (jset.com_out[0] == 0) strcpy(jset.com_out,  "/dev/null");

	fd_comin = open(jset.com_in,            O_RDONLY          );
	fd_comout= open(jset.com_out, O_CREAT | O_WRONLY | O_APPEND | O_NDELAY);

        tcgetattr(fd_comin, &t);
        cfmakeraw(&t);
        tcsetattr(fd_comin, TCSADRAIN, &t);

        tcgetattr(fd_comout, &t);
        cfmakeraw(&t);
        tcsetattr(fd_comout, TCSADRAIN, &t);

	
	setspeed(spd);
}


void set_com(void)
{
	com_deinit();
	com_init();
}

