/** JOYCE PCW emulator for DJGPP *****************************/
/**                                                         **/
/** JOYCEMEM: Memory manager                                **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1998                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"

/* The idea of the paging is that PCWRD[addr>>14][addr] will give the 
  address of the byte at addr. Separate read and write pointers are 
  maintained because of the CPC-like paging mode (which AFAIK is not 
  used by any PCW software). */

static unsigned char last_mpage[4] = { 0x80, 0x81, 0x82, 0x83 };

static int nbanks;

/* Allocate the memory */

int mem_alloc(void)
{
	long siz=2097152;        /* 2048k, ie 2 megs */

	if (PCWRAM) return (nbanks*16);

	while (siz>262144)  /* ie 256k */
	{
		PCWRAM=malloc(siz);
		if (PCWRAM)
		{
			nbanks=siz/16384;
			memset(PCWRAM, 0, siz);
			return(siz/1024);
		}
		siz-=262144;
	}
	return(0);
}


/* Process an OUT 0xF4. If one of the top 4 bits is set, memory reads and
 * writes will be to the same bank. This is only ever a concern if the 
 * CPC-like paging system is used; according to Richard Clayton CP/M and
 * LocoScript do not use this system.
 */

void mem_lock(unsigned char value)
{
	int np;

	bank_lock[0] = value & 0x40;
	bank_lock[1] = value & 0x10;
	bank_lock[2] = value & 0x20;
	bank_lock[3] = value & 0x80;

	/* If memory was paged the CPC way, then paging may have to be
          recalculated */

	for(np = 0; np < 4; np++)
	{
		if (!(last_mpage[np] & 0x80)) mem_page(np, last_mpage[np]);
	}
}

		
void mem_page(unsigned char port, unsigned char value)
{
	int bankoff;

	port &=3;               /* This handles the memory paging. */
	last_mpage[port] = value;
	if (value & 0x80)	/* `Extended' (PCW-like) paging */
	{
		value = value % nbanks;      /* The idea is that PCWRD[addr>>14] */
		bankoff=(value-port)*16384;  /* +addr gives the correct address */
		PCWWR[port] = PCWRD[port] = PCWRAM+bankoff;
	}
	else
	{
		int rv, wv;
		rv = (value >> 4) & 7;
		wv = value & 7;
		bankoff=(rv-port)*16384;  /* +addr gives the correct address */
		PCWRD[port] = PCWRAM+bankoff;
		bankoff=(wv-port)*16384;  /* +addr gives the correct address */
		PCWWR[port] = PCWRAM+bankoff;
		if (bank_lock[port]) PCWRD[port] = PCWWR[port];
	}
	
}


/* Read and write the PCW's memory - inlined
 *
 *
 *byte RdZ80(word addr)
 *{
 *	return (PCWRD[addr>>14])[addr];
 *}
 *
 *
 *void WrZ80(word addr,byte v)
 *{
 *	(PCWWR[addr>>14])[addr]=v;
 *}
 *
 */
