/** JOYCE PCW emulator for DJGPP *****************************/
/**                                                         **/
/** JOYCEPTR: Emulates Centronics printer                   **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1998                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "Joyceptr.h"

void PtrSettings(void)	/* Change printer settings.  */
{
#ifndef __MSDOS__

	if (!(jset.cen_port & 0x80))
	{
		sprintf(jset.printer_file, "/dev/lp%d", jset.cen_port);
		jset.cen_port |= 0x80;
	}

#endif
	if ((jset.cen_port & 0x80) && !printer_fp)	/* switching TO file */
	{
		/* Note opening in append mode */

		printer_fp = fopen(jset.printer_file, "a");
	}
	
	if (!(jset.cen_port & 0x80) && printer_fp)	/* switching FROM file */
	{
		fclose(printer_fp);
		printer_fp = NULL;
	}
	
}



void InitPtr(void)	/* Initialise printer. */
{
	if ((jset.cen_port & 0x80) && !printer_fp)	/* switching TO file */
	{
		/* Note opening in append mode */

		printer_fp = fopen(jset.printer_file, "a");
	}
}



void UnitPtr(void)
{
	if (printer_fp)
	{
		fclose(printer_fp);
		printer_fp = NULL;
	}
}


void cen_out(byte port, byte value) /* Byte written to a CEN port */
{
	static byte lastch = 0;

	if (jset.cen_port & 0x80) switch(port)
	{
		case 0x85:
		if (printer_fp) fputc(lastch, printer_fp);		
		break;

		case 0x87:
		lastch = value;
		break;
	}
#ifdef __MSDOS__
	else switch(port)
	{
		case 0x85:	/* Toggle STROBE */
		_bios_printer(_PRINTER_WRITE, jset.cen_port, lastch);
		break;

		case 0x86:	/* Reset the port */
		_bios_printer(_PRINTER_INIT, jset.cen_port, 0);
		break;

		case 0x87:	/* Sending character */
		lastch = value;
		break;
	}
#endif
}

byte cen_in(byte port)	/* Byte read from CEN port */
{
	/* v1.11: Text file? */

	if (jset.cen_port & 0x80) switch (port)
	{
		case 0x84:
		case 0x85:
		return (printer_fp ? 0xFE : 0xFF);

		default:
		return 0xFF;
	}
#ifdef __MSDOS__
	else switch(port)
	{
		byte ch;

		case 0x84:	/* BUSY? */
		ch = _bios_printer(_PRINTER_STATUS, jset.cen_port, 0) & 0x80;
		if (ch & 0x80) return 0xFE; 
		else	       return 0xFF;				
		break;

		case 0x85:	/* Detect CEN: */
		return 0xFE;
		default:	/* Detect CEN: */
		return 0xFF;
	}
#else
	return 0xFF;
#endif
}		



void list(Z80 *R)	/* (v1.26) Printer bypass support */
{
	if (jset.cen_port & 0x80) 
	{
		if (printer_fp) fputc(R->BC.B.l, printer_fp);		
	}
#ifdef __MSDOS__
	else
	{
		_bios_printer(_PRINTER_WRITE, jset.cen_port, R->BC.B.l);
	}
#endif
}


void listst(Z80 *R)	/* (v1.26) Printer bypass support */
{
	if (jset.cen_port & 0x80) /* Printing to a file */
	{
		R->AF.B.h = (printer_fp ? 1 : 0);
		return;
	}
#ifdef __MSDOS__
	R->AF.B.h = (_bios_printer(_PRINTER_STATUS, jset.cen_port, 0) & 0x80) ? 1 : 0;
#endif
}

/* (no MATRIX printer) 

void matrix_out(byte port, byte value);
byte matrix_in(byte port);

*/


