/** JOYCE PCW emulator for DJGPP *****************************/
/**                                                         **/
/** JOYCESET.C - read/write settings                        **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1998                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "Joyceset.h"

int read_settings(void)
{
	FILE *fp;
	char ini_path[200];
	int rcnt;
	byte *jsmem = (byte *)&jset1;

	strcpy(ini_path, find_joyce_file("joyce.cfg"));
	fp = fopen(ini_path,"rb");
	if (!fp) return 0;
	rcnt = fread(&jset1, 1, sizeof(JOYCE_SETTINGS), fp);
	fclose(fp);

	/* The structure grows between versions. We grow the old one up to
	 * the new one by filling it with zeroes
	 */

	while (rcnt < sizeof(JOYCE_SETTINGS)) jsmem[rcnt++] = 0;

	/* (v1.23) If mouse ballistics are zero, set them to defaults */

	if (!jset1.mxscale) jset1.mxscale = 4;
	if (!jset1.myscale) jset1.myscale = 2;
	if (!jset1.mxspeed) jset1.mxspeed = 15;
	if (!jset1.myspeed) jset1.myspeed = 15;

	if (memcmp(jset1.magic, SETTINGS_MAGIC, sizeof(jset.magic))) return 0;
	memcpy(&jset, &jset1, sizeof(JOYCE_SETTINGS));
	
	/* Update settings */

	PtrSettings();
	return 1;
}

int write_settings(void)
{
	FILE *fp;
	char ini_path[200];
	int rcnt;

	memcpy(&jset1, &jset, sizeof(JOYCE_SETTINGS));
	memcpy(jset1.magic, SETTINGS_MAGIC, sizeof(jset1.magic));

        strcpy(ini_path, find_joyce_file("joyce.cfg"));
	fp = fopen(ini_path,"wb");
	if (!fp) return 0;
	rcnt = fwrite(&jset1, sizeof(JOYCE_SETTINGS), 1, fp);
	fclose(fp);
	if (rcnt < 1) return 0;
	return 1;
}


void default_settings(void)
{
	memset(&jset, 0, sizeof(jset));
	memcpy(jset.magic, SETTINGS_MAGIC, sizeof(jset.magic));
	jset.screen_refresh = 144;
	jset.verbose = 1;
}
