/** JOYCE PCW emulator ***************************************/
/**                                                         **/
/** MD3TRANS - converts between discs and images. Note:     **/
/**                                                         **/
/** 1. This is a 16-bit program, to be compiled with        **/
/** with Pacific C, rather than a 32-bit DJGPP program.     **/
/**                                                         **/
/** 2. The DOSBIOS.OBJ file contains simple wrappers for    **/
/**   the BIOS calls. Unfortunately, I've lost the source   **/
/**   for this module.                                      **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1999                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "dosbios.h"
#include <dos.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <conio.h>
#include <string.h>
#ifndef byte
typedef unsigned char byte;
#endif


typedef unsigned long  ulong;
typedef unsigned short ushort;
typedef unsigned char far * ufar; 

byte trkbuf[8192];	/* Buffers for constructing image tracks */
byte secbuf[8192];	/* or loading them */
byte specbuf[512];	/* Disc specification sector */
byte formbuf[128];	/* Parameters for the format command */

#pragma pack(1)		/* byte-aligned structs */
		
typedef struct		/* CPCEMU disc image header */
{
	char   magic[0x30];	/* "MV-CPC" */
	byte   ntrks;		/* no. of tracks */
	byte   nhds;		/* no. of heads */
	ushort trklen;		/* size of a track */
	byte   spare[0xCC];
	
} CPCHDR;


typedef struct		/* CPCEMU track header */
{
	char  magic[0x10];	/* "Track-Info" */
	byte  trkno;		/* Cylinder no. */
	byte  trkhead;		/* Head no. */
	byte  unused[2];
	byte  BPS;		/* bytes/sector */
	byte  SPT;		/* sectors/track */
	byte  GAP3;		/* gap value */
	byte  FILLER;		/* format filler byte */
	byte  secinfo[0xE8];	/* sector info records */
} CPCTRK;

CPCHDR header = {"MV - CPCEMU Disk-File\r\nDisk-Info\r\n(JOYCE)",
				 80,	/* cylinders */
				 2,	/* heads */
				 4864,	/* track length */
				 {0}};

CPCTRK trackinf = {"Track-Info\r\n", 
				   0,	 /* Track no. */
				   0,	 /* Head */
				   {0},	 /* Unused */
				   2,	 /* 512 bps */
				   9,	 /* SPT */
				   0x4E, /* Gap */
				   0xE5, /* Filler */
				   {0}};	

/* Format a track. This routine is very similar to the ones in the lost
  DOSBIOS module */

byte format_trk(byte drive, byte hd, byte trk, byte spt, byte far *buf)
{
	int retries = 3;
	int err = 1;
	union REGS ir, or;
	struct SREGS segs;

	ir.x.ax = 0x1704;	/* 3.5" disc in 3.5" drive */
	ir.h.dl = drive;
	int86(0x13, &ir, &or);
	
	while (err)
	{
		--retries;
		if (!retries) return err;

		ir.h.ah = 5;
		ir.h.al = spt;
		ir.h.ch = trk;
		ir.h.dh = hd;
		ir.h.dl = drive;
		ir.x.bx = FP_OFF(buf);
		segs.es = FP_SEG(buf);
		int86x(0x13,&ir, &or, &segs);

		err = or.h.ah;
	}
	return err;
}

/* Report an error, ask what to do */

char diskerr(byte err, char drive)
{
	char action=0;
	
	switch(err)
	{
		case 1:
		fprintf(stderr,"Drive %c: is not a valid floppy drive\n",drive);
		exit(2);
		break;
		case 2:
		fprintf(stderr,"Drive %c: Missing address mark",drive);
		break;
		case 3:
		fprintf(stderr,"Drive %c: Disc read-only",drive);
		break;
		case 4:
		fprintf(stderr,"Drive %c: Sector not found",drive);
		break;
		case 6:
		return 'R';	//fprintf(stderr,"Drive %c: Disc changed",drive);
					//break;
		case 8:
		fprintf(stderr,"DMA overrun");
		break;
		case 9:
		fprintf(stderr,"DMA over segment boundary\n");
		exit(4);
		break;
		case 16:
		fprintf(stderr,"Drive %c: CRC data error",drive);
		break;
		case 32:
		fprintf(stderr,"Floppy disc controller error");
		break;
		case 64:
		fprintf(stderr,"Drive %c: Seek fail",drive);
		break;
		case 128:
		fprintf(stderr,"Drive %c: not ready",drive);
		break;
		default:
		fprintf(stderr,"Drive %c unknown error",drive);
	}
	fprintf(stderr," - Retry, Ignore or Cancel?");
	fflush(stderr);
	
	while (action==0)
		{
		action=getch();
		if (action=='C' || action=='c') exit(4);
		else if ((action!='R') && (action !='r') && 
		         (action!='I') && (action !='i')) action=0;
		}
	fputc('\n',stderr);
	if (islower(action)) action = toupper(action);
	return(action);
}

/* Select our disc parameter table */

ufar disctable(ufar newtable)
{
	ufar far *tbl = (ufar far *)0x78;	/* The disc parameters table */
	ufar oldv;

	oldv = *tbl;
	*tbl = newtable;
	
	return oldv;

}




/* Image to drive copy */

void do_rtran(char dstdrive, char *discfile)
{
	static byte mytbl[11]; 
	ufar oldtbl;
	byte drive = 0;
	byte err = 0;
	byte trk, hd, sec;
	CPCTRK *tptr = (CPCTRK *)trkbuf;
	
	char act = 'R';

	FILE *img;

	img = fopen(discfile,"rb");
	if (!img 
	|| (fread(&header, 1, 256, img) < 256) 
	||  memcmp(header.magic,"MV - CPC",8))
	{
		if (img) fclose(img);
		fprintf(stderr, "Cannot open or invalid format: file %s\n", discfile);
		return;
	}
	
	oldtbl = disctable(mytbl);
	disctable(oldtbl);
	
	for (sec = 0; sec < 11; sec++) mytbl[sec] = oldtbl[sec];
										/* Copy DOS's DPT to ours. */
	
	printf("\n\nCopying from %s to %c:\n\n", discfile, dstdrive);
	
	drive = dstdrive - 'A';

	disctable(mytbl);
	dskreset();
	disctable(oldtbl);

	{
		union REGS rg;

		rg.x.ax = 0x1704;	/* Select 3.5" disc in 3.5" drive */
		rg.h.dl = drive;
		int86(0x13,&rg,&rg);
	}
	

	for (trk = 0; trk < header.ntrks; trk++)
		for (hd = 0; hd < header.nhds; hd++)
		{
			if ((fread(trkbuf, 1, header.trklen, img) < header.trklen)
			|| !memcmp(trkbuf, "Track-Info\r\n", 14))
			{
				fprintf(stderr,"\nNot a valid DSK image: %s\n", discfile);
				fclose(img);
				return;	
			}
			for (sec = 0; sec < tptr->SPT; sec++)
			{
				formbuf[sec*4 + 0] = tptr->secinfo[sec*8 + 0];
				formbuf[sec*4 + 1] = tptr->secinfo[sec*8 + 1];
				formbuf[sec*4 + 2] = tptr->secinfo[sec*8 + 2];
				formbuf[sec*4 + 3] = tptr->secinfo[sec*8 + 3];
			}
			mytbl[4] = tptr->SPT;
			mytbl[8] = 0xe5;
			if (tptr->SPT == 10) mytbl[7] = 0x17;
			act = 'R';
			while (act != 'I')
			{
				disctable(mytbl);
				err = format_trk(drive, hd, trk, tptr->SPT, formbuf);
				disctable(oldtbl);
		
				if (!err) act = 'I';
				else 
				{
					printf("\n");
					act = diskerr(err, dstdrive);
				}
			}
			
			for (sec = 0; sec < tptr->SPT; sec++)
			{

				printf("Cylinder %03d/%03d   Head  %d/%d   Sector  %02d/%02d\r",
						trk + 1, header.ntrks, 
						hd + 1,  header.nhds,
						sec + 1, tptr->SPT);
				fflush(stdout);
			
				act = 'R';
				while (act != 'I')
				{
					disctable(mytbl);
					err = dskwrite(drive, hd, trk, 
								   tptr->secinfo[sec * 8 + 2], 1, 
								   trkbuf + sec*0x200 + 0x100);
					disctable(oldtbl);
		
					if (!err) act = 'I';
					else 
					{
						printf("\n");
						act = diskerr(err, dstdrive);
					}
				}
				
			}
		}

	fclose(img);
	printf("\nDisc complete.\n");

	
}


/* Drive to image copy */

void do_trans(char srcdrive, char *discfile)
{
	static byte mytbl[11]; 
	ufar oldtbl;
	byte drive = 0;
	byte err = 0;
	byte trk, hd, sec;

	int seclen;
	char act = 'R';

	FILE *img;

	img = fopen(discfile,"wb");
	if (!img)
	{
		fprintf(stderr, "Cannot create disc image file %s\n", discfile);
		return;
	}
	
	oldtbl = disctable(mytbl);
	disctable(oldtbl);
	
	for (sec = 0; sec < 11; sec++) mytbl[sec] = oldtbl[sec];
										/* Copy DOS's DPT to ours. */
	
	printf("\n\nCopying from %c: to %s\n\n", srcdrive, discfile);
	
	drive = srcdrive - 'A';
	act = 'R';
	while (act != 'I')
	{
		disctable(mytbl);
		dskreset();
		err = dskread(drive, 0, 0, 1, 1, specbuf);
		disctable(oldtbl);
		
		if (!err) act = 'I';
		else act = diskerr(err, srcdrive);
	}

	header.ntrks  = specbuf[2];
	header.nhds   = (specbuf[1] & 3) ? 2 : 1;
	header.trklen = 0x100 + (specbuf[3] * 0x200);

	if ((header.ntrks < 40)
	||  (header.ntrks > 84) 
	||  (header.trklen < 4352)
	||  (header.trklen > 5376))
	{
		fprintf(stderr,"Can't identify the format of this disc.\n");
		fclose(img);
		remove(discfile);
		return;
	} 

	if (fwrite(&header, 1, 256, img) < 256)
	{
		fprintf(stderr,"Disc write error\n");
		fclose(img);
		remove(discfile);
		return;
	}

	mytbl[3] = specbuf[4];	/* Bytes / sector */
	mytbl[4] = specbuf[3];	/* Sectors / track */
	mytbl[5] = specbuf[8];	/* Gap length */
	seclen   = (128 << specbuf[4]);

	disctable(mytbl);
	dskreset();
	disctable(oldtbl);
	
	for (trk = 0; trk < header.ntrks; trk++)
		for (hd = 0; hd < header.nhds; hd++)
		{
			/* Special for MD3 */
		
			if (trk == 1 && hd == 1)
			{
				specbuf[4] = 1; /* 256 bytes/sector */				
				mytbl[3] = 1;
				seclen   = (128 << specbuf[4]);
			}
			else
			{
				specbuf[4] = 2; /* 512 bytes/sector */
				mytbl[3] = 2;
				seclen   = (128 << specbuf[4]);
			}

			/* end special */
			trackinf.trkno   = trk;
			trackinf.trkhead = hd;
			trackinf.BPS	 = specbuf[4];
			trackinf.SPT	 = specbuf[3];
			trackinf.GAP3	 = 0x4E;
			trackinf.FILLER  = 0xE5;
			for (sec = 0; sec < specbuf[3]; sec++)
			{
				trackinf.secinfo[0x00 + 8*sec] = trk;
				trackinf.secinfo[0x01 + 8*sec] = hd;
				trackinf.secinfo[0x02 + 8*sec] = sec + 1;
				trackinf.secinfo[0x03 + 8*sec] = specbuf[4];
				trackinf.secinfo[0x04 + 8*sec] =
				trackinf.secinfo[0x05 + 8*sec] =
				trackinf.secinfo[0x06 + 8*sec] =
				trackinf.secinfo[0x07 + 8*sec] = 0;
			}
			if (fwrite(&trackinf, 1, 256, img) < 256)
			{
				fprintf(stderr,"Disc write error\n");
				fclose(img);
				remove(discfile);
				return;
			}
			for (sec = 0; sec < specbuf[3]; sec++)
			{

				printf("Cylinder %03d/%03d   Head  %d/%d   Sector  %02d/%02d\r",
						trk + 1, header.ntrks, 
						hd + 1,  header.nhds,
						sec + 1, specbuf[3]);
				fflush(stdout);
			
				act = 'R';
				while (act != 'I')
				{
					disctable(mytbl);
					err = dskread(drive, hd, trk, sec + 1, 1, secbuf);
					disctable(oldtbl);
		
					if (!err) act = 'I';
					else 
					{
						printf("\n");
						act = diskerr(err, srcdrive);
					}
				}
				if (fwrite(secbuf, 1, seclen, img) < seclen)
				{
					fprintf(stderr,"\nDisc write error\n");
					fclose(img);
					remove(discfile);
					return;
				}
			}
			/* Special for MD3: pad out the track, so that it occupies */
			/* the same space in the image as the others */
			if (trk == 1 && hd == 1)
			{
				for (sec = 0; sec < specbuf[3]; sec++)
				{
					if (fwrite(secbuf, 1, seclen, img) < seclen)
					{
						fprintf(stderr,"\nDisc write error\n");
						fclose(img);
						remove(discfile);
						return;
					}
				}
			}
			/* end special */	
		}

	fclose(img);
	printf("\nDisc complete.\n");

	
}






void help(void)
{
	printf("MD3TRANS v1.00 - Convert Microdesign boot disc -> image file\n\n");

	printf("Syntax: MD3TRANS x: dskfile\n\n");

	printf("x: must be a floppy drive - A: or B:\n");
	exit(1);
}



void main (int argc, char **argv)
{
	char drv[2];
	
	if (argc < 3) help();

	drv[0] = argv[1][0];
	drv[1] = argv[2][0];
	
	if (islower(drv[0])) drv[0] = argv[1][0] = toupper(drv[0]);
	if (islower(drv[1])) drv[1] = argv[2][0] = toupper(drv[1]);

	if      (!strcmp(argv[1], "A:") || !strcmp(argv[1], "B:")) do_trans(drv[0], argv[2]);
//	else if (!strcmp(argv[2], "B:") || !strcmp(argv[2], "A:")) do_rtran(drv[1], argv[1]); 
	else help();
	
}
	
